"""
AST to AST transformation

The preprocessing pass replaces all _candidate_ dataflow regions with
calls to PaSh's runtime to let it establish if they are actually dataflow
regions. The pass serializes all candidate dataflow regions:
- A list of ASTs if at the top level or
- an AST subtree if at a lower level

The PaSh runtime then deserializes them, compiles them (if safe) and optimizes them.
"""

from shasta.ast_node import AstNode

from shell_ast.ast_util import PreprocessedAST, UnparsedScript, unzip
from shell_ast.walk_preprocess import WalkPreprocess, PreprocessContext
from shell_ast.handlers.loop_tracking import for_node_with_loop_tracking
from shell_ast.transformation_options import AbstractTransformationState


# === Walker setup ===


def create_pash_walker() -> WalkPreprocess:
    """
    Create a preprocessing walker configured for PaSh.

    Returns:
        WalkPreprocess instance with PaSh-specific handlers
    """
    handlers = {
        "for": for_node_with_loop_tracking,
    }
    return WalkPreprocess(handlers=handlers)


# Module-level walker instance
_pash_walker = create_pash_walker()


def preprocess_node(
    ast_node: AstNode,
    trans_options: AbstractTransformationState,
    last_object: bool,
) -> PreprocessedAST:
    """
    Preprocesses an AstNode. Given an AstNode of any type, it will appropriately
    dispatch a preprocessor for the specific node type.

    This is the main entry point for preprocessing.

    Parameters:
        ast_node (AstNode): The AstNode to parse
        trans_options (AbstractTransformationState):
            A concrete transformation state instance corresponding to the output target
        last_object (bool): Flag for whether this is the last AstNode

    Returns:
        PreprocessedAst: the preprocessed version of the original AstNode
    """
    ctx = PreprocessContext(trans_options=trans_options, last_object=last_object)
    return _pash_walker.walk(ast_node, ctx)


# === Main transformation function ===


def replace_ast_regions(ast_objects, trans_options: AbstractTransformationState):
    """
    Replace candidate dataflow AST regions with calls to PaSh's runtime.
    """
    preprocessed_asts = []
    candidate_dataflow_region = []
    last_object = False
    for i, ast_object in enumerate(ast_objects):
        # If we are working on the last object we need to keep that in mind when replacing.
        # The last df-region should not be executed in parallel no matter what (to not lose its exit code.)
        if i == len(ast_objects) - 1:
            last_object = True

        ast, original_text, _linno_before, _linno_after = ast_object
        assert isinstance(ast, AstNode)

        # Preprocess ast by replacing subtrees with calls to runtime.
        # - If the whole AST needs to be replaced (e.g. if it is a pipeline)
        #   then the second output is true.
        # - If the next AST needs to be replaced too (e.g. if the current one is a background)
        #   then the third output is true
        preprocessed_ast_object = preprocess_node(
            ast, trans_options, last_object=last_object
        )
        # If the dataflow region is not maximal then it implies that the whole
        # AST should be replaced.
        assert (
            not preprocessed_ast_object.is_non_maximal()
            or preprocessed_ast_object.should_replace_whole_ast()
        )

        # If the whole AST needs to be replaced then it implies that
        # something will be replaced
        assert (
            not preprocessed_ast_object.should_replace_whole_ast()
            or preprocessed_ast_object.will_anything_be_replaced()
        )

        # If it isn't maximal then we just add it to the candidate
        if preprocessed_ast_object.is_non_maximal():
            candidate_dataflow_region.append(
                (preprocessed_ast_object.ast, original_text)
            )
        else:
            # If the current candidate dataflow region is non-empty
            # it means that the previous AST was in the background so
            # the current one has to be included in the process no matter what
            if len(candidate_dataflow_region) > 0:
                candidate_dataflow_region.append(
                    (preprocessed_ast_object.ast, original_text)
                )
                # Since the current one is maximal (or not wholy replaced)
                # we close the candidate.
                dataflow_region_asts, dataflow_region_lines = unzip(
                    candidate_dataflow_region
                )
                dataflow_region_text = join_original_text_lines(dataflow_region_lines)
                replaced_ast = trans_options.replace_df_region(
                    dataflow_region_asts,
                    ast_text=dataflow_region_text,
                    disable_parallel_pipelines=last_object,
                )
                candidate_dataflow_region = []
                preprocessed_asts.append(replaced_ast)
            else:
                if preprocessed_ast_object.should_replace_whole_ast():
                    replaced_ast = trans_options.replace_df_region(
                        [preprocessed_ast_object.ast],
                        ast_text=original_text,
                        disable_parallel_pipelines=last_object,
                    )
                    preprocessed_asts.append(replaced_ast)
                else:
                    # In this case, it is possible that no replacement happened,
                    # meaning that we can simply return the original parsed text as it was.
                    if (
                        preprocessed_ast_object.will_anything_be_replaced()
                        or original_text is None
                    ):
                        preprocessed_asts.append(preprocessed_ast_object.ast)
                    else:
                        preprocessed_asts.append(UnparsedScript(original_text))

    # Close the final dataflow region
    if len(candidate_dataflow_region) > 0:
        dataflow_region_asts, dataflow_region_lines = unzip(candidate_dataflow_region)
        dataflow_region_text = join_original_text_lines(dataflow_region_lines)
        replaced_ast = trans_options.replace_df_region(
            dataflow_region_asts,
            ast_text=dataflow_region_text,
            disable_parallel_pipelines=True,
        )
        candidate_dataflow_region = []
        preprocessed_asts.append(replaced_ast)

    return preprocessed_asts


def join_original_text_lines(shell_source_lines_or_none):
    """
    Join original unparsed shell source in a safe way,
    handling the case where some of the text is None (e.g., in case of stdin parsing).
    """
    if any([text_or_none is None for text_or_none in shell_source_lines_or_none]):
        return None
    else:
        return "\n".join(shell_source_lines_or_none)



from pash_annotations.datatypes.BasicDatatypes import Flag, ArgStringType, Operand
from pash_annotations.datatypes.BasicDatatypesWithIO import OptionWithIO
from pash_annotations.datatypes.CommandInvocationInitial import CommandInvocationInitial
from pash_annotations.annotation_generation.datatypes.InputOutputInfo import (
    InputOutputInfo,
)
from pash_annotations.annotation_generation.datatypes.ParallelizabilityInfo import (
    ParallelizabilityInfo,
)
from pash_annotations.annotation_generation.datatypes.CommandProperties import (
    CommandProperties,
)
from pash_annotations.annotation_generation.AnnotationGeneration import (
    get_input_output_info_from_cmd_invocation,
    get_parallelizability_info_from_cmd_invocation,
)
from pash_annotations.datatypes.CommandInvocationWithIOVars import (
    CommandInvocationWithIOVars,
)

from ir_defs.arg import Arg

# for typing
from pash_annotations.datatypes.CommandInvocationPrefix import CommandInvocationPrefix

from util import (
    string_to_argument,
    redir_stdout_to_file,
    redir_file_to_stdin,
    make_command,
)


def get_command_invocation_prefix_from_dfg_node(dfg_node):
    return CommandInvocationPrefix(
        cmd_name=dfg_node.com_name,
        flag_option_list=dfg_node.flag_option_list,
        positional_config_list=dfg_node.positional_config_list,
    )


# TODO: ideally methods in the respective classes but requires refactoring of parsing infrastructure
# TODO: isn't this `to_ast`?
def to_node_cmd_inv_with_io_vars(cmd_inv, edges, redirs, assignments):
    ast_cmd_name = string_to_argument(cmd_inv.cmd_name)
    ast_flagoptions = []
    for flagoption in cmd_inv.flag_option_list:
        ast_flagoptions += to_ast_flagoption(flagoption, edges)
    ast_operands = [to_ast_operand(operand, edges) for operand in cmd_inv.operand_list]
    cmd_asts = [ast_cmd_name] + ast_flagoptions + ast_operands

    # TODO: check for actual stdin
    stdin_redir = []
    if cmd_inv.implicit_use_of_streaming_input is not None:
        fid, _, _ = edges[cmd_inv.implicit_use_of_streaming_input]
        if not (fid.has_file_descriptor_resource() and fid.resource.is_stdin()):
            stdin_redir = [redir_file_to_stdin(fid.to_ast())]

    # TODO: check for actual stdout
    stdout_redir = []
    if cmd_inv.implicit_use_of_streaming_output is not None:
        fid, _, _ = edges[cmd_inv.implicit_use_of_streaming_output]
        if not (fid.has_file_descriptor_resource() and fid.resource.is_stdout()):
            stdout_redir = [redir_stdout_to_file(fid.to_ast())]

    new_redirs = redirs + stdin_redir + stdout_redir
    node = make_command(cmd_asts, redirections=new_redirs, assignments=assignments)
    return node


def to_ast_flagoption(flagoption, edges):
    if isinstance(flagoption, Flag):
        return [string_to_argument(flagoption.get_name())]
    elif isinstance(flagoption, OptionWithIO):  # retype to IOVar
        opt_name_ast = string_to_argument(flagoption.get_name())
        opt_arg_ast = translate_io_var_if_applicable(flagoption.get_arg(), edges)
        return [opt_name_ast, opt_arg_ast]


def to_ast_operand(operand, edges):
    if isinstance(operand, Operand):
        return translate_io_var_if_applicable(operand.get_name(), edges)
    return translate_io_var_if_applicable(operand, edges)


def translate_io_var_if_applicable(pot_io_var, edges):
    # TODO: this is currently a hack but eventually every possible type gets their own to_ast-function
    if isinstance(pot_io_var, int):
        return dereference_io_var(pot_io_var, edges)
    elif isinstance(pot_io_var, ArgStringType):
        return to_ast_arg_string_type(pot_io_var)
    elif isinstance(pot_io_var, CommandInvocationWithIOVars):
        assert False
        # only happens as r-wrapped node
        return to_node_cmd_inv_with_io_vars(pot_io_var, edges, [], [])
    elif isinstance(pot_io_var, Arg):
        return pot_io_var.to_ast()
    else:
        raise Exception("Unhandled type for operand in to_ast!")


def to_ast_arg_string_type(arg_string_type):
    return arg_string_type.get_name().arg_char_list  # is of type Arg


# assumes io_var is an edge id
def dereference_io_var(io_var, edges):
    fid, _, _ = edges[io_var]
    return fid.to_ast()


def get_input_output_info_from_cmd_invocation_util(
    cmd_invocationInitial: CommandInvocationInitial,
) -> InputOutputInfo:
    return get_input_output_info_from_cmd_invocation(cmd_invocationInitial)


def get_parallelizability_info_from_cmd_invocation_util(
    cmd_invocationInitial: CommandInvocationInitial,
) -> ParallelizabilityInfo:
    return get_parallelizability_info_from_cmd_invocation(cmd_invocationInitial)


def construct_property_container_from_list_of_properties(list_properties):
    return CommandProperties(dict(list_properties))


# this function is needed to wrap a node in `r_wrap`
def to_arg_from_cmd_inv_with_io_vars_without_streaming_inputs_or_outputs_for_wrapping(
    cmd_inv, edges
):
    # we already expand here
    whole_cmd = Arg.string_to_arg("'")
    arg_cmd_name = Arg.string_to_arg(cmd_inv.cmd_name)
    arg_flagoptions = []
    for flagoption in cmd_inv.flag_option_list:
        arg_flagoptions += to_arg_flagoption(flagoption, edges)
    arg_operands = [to_arg_operand(operand, edges) for operand in cmd_inv.operand_list]
    all_cmd_parts_arg = [arg_cmd_name]
    all_cmd_parts_arg.extend(arg_flagoptions)
    all_cmd_parts_arg.extend(arg_operands)
    for part in all_cmd_parts_arg:
        whole_cmd.concatenate(part)
    whole_cmd.concatenate(Arg.string_to_arg("'"))
    return whole_cmd


def to_arg_flagoption(flagoption, edges):
    if isinstance(flagoption, Flag):
        return [Arg.string_to_arg(flagoption.get_name())]
    elif isinstance(flagoption, OptionWithIO):
        opt_name_arg = Arg.string_to_arg(flagoption.get_name())
        opt_arg_arg = translate_io_var_to_arg_if_applicable(flagoption.get_arg(), edges)
        return [opt_name_arg, opt_arg_arg]


def to_arg_operand(operand, edges):
    if isinstance(operand, Operand):
        return translate_io_var_to_arg_if_applicable(operand.get_name(), edges)
    return translate_io_var_to_arg_if_applicable(operand, edges)


def translate_io_var_to_arg_if_applicable(pot_io_var, edges):
    if isinstance(pot_io_var, int):
        return Arg(dereference_io_var(pot_io_var, edges))
    elif isinstance(pot_io_var, ArgStringType):
        result = pot_io_var.get_name()  # is of type Arg
        return result
    else:
        raise Exception("Unhandled type for operand in to_arg!")



"""
Shell AST manipulation for preprocessing.
"""

from shell_ast.ast_to_ast import preprocess_node, replace_ast_regions
from shell_ast.transformation_options import (
    TransformationType,
    TransformationState,
    SpeculativeTransformationState,
    AirflowTransformationState,
)

__all__ = [
    "preprocess_node",
    "replace_ast_regions",
    "TransformationType",
    "TransformationState",
    "SpeculativeTransformationState",
    "AirflowTransformationState",
]



"""
Utility functions for the speculative execution component.
"""

import os

from util import log, PASH_TMP_PREFIX


def initialize(trans_options) -> None:
    """Initialize the partial order directory."""
    dir_path = partial_order_directory()
    os.makedirs(dir_path)


def partial_order_directory() -> str:
    """Return the path to the partial order directory."""
    return f"{PASH_TMP_PREFIX}/speculative/partial_order/"


def partial_order_file_path():
    """Return the path to the partial order file."""
    return f"{PASH_TMP_PREFIX}/speculative/partial_order_file"


def initialize_po_file(trans_options, dir_path) -> None:
    """Initialize the partial order file."""
    with open(trans_options.get_partial_order_file(), "w") as f:
        f.write(f"# Partial order files path:\n")
        f.write(f"{dir_path}\n")


def scheduler_server_init_po_msg(partial_order_file: str) -> str:
    """Create message to initialize scheduler with partial order file."""
    return f"Init:{partial_order_file}"


def save_df_region(
    text_to_output: str, trans_options, df_region_id: int, predecessor_ids: int
) -> None:
    """Save a dataflow region to a file."""
    # Associate nodes with their surrounding loops
    current_loop_context = trans_options.get_current_loop_context()
    log("Df region:", df_region_id, "loop context:", current_loop_context)

    # Add the loop context to the partial_order state
    trans_options.add_node_loop_context(df_region_id, current_loop_context)

    # Save df_region as text in its own file
    df_region_path = f"{partial_order_directory()}/{df_region_id}"
    with open(df_region_path, "w", encoding="utf-8") as f:
        f.write(text_to_output)

    # Save the edges in the partial order state
    for predecessor in predecessor_ids:
        trans_options.add_edge(predecessor, df_region_id)


def serialize_edge(from_id: int, to_id: int) -> str:
    """Serialize an edge in the partial order."""
    return f"{from_id} -> {to_id}\n"


def serialize_number_of_nodes(number_of_ids: int) -> str:
    """Serialize the number of nodes."""
    return f"{number_of_ids}\n"


def serialize_loop_context(node_id: int, loop_contexts) -> str:
    """Serialize a loop context."""
    loop_contexts_str = ",".join([str(loop_ctx) for loop_ctx in loop_contexts])
    return f"{node_id}-loop_ctx-{loop_contexts_str}\n"


def save_number_of_nodes(trans_options):
    """Save the number of nodes to the partial order file."""
    number_of_ids = trans_options.get_number_of_ids()
    po_file_path = trans_options.get_partial_order_file()
    with open(po_file_path, "a") as po_file:
        po_file.write(serialize_number_of_nodes(number_of_ids))


def save_loop_contexts(trans_options):
    """Save loop contexts to the partial order file."""
    loop_context_dict = trans_options.get_all_loop_contexts()
    log("Loop context dict:", loop_context_dict)
    po_file_path = trans_options.get_partial_order_file()
    with open(po_file_path, "a") as po_file:
        for node_id in sorted(loop_context_dict.keys()):
            loop_ctx = loop_context_dict[node_id]
            po_file.write(serialize_loop_context(node_id, loop_ctx))


def serialize_partial_order(trans_options):
    """Serialize the complete partial order to a file."""
    # Initialize the po file
    dir_path = partial_order_directory()
    initialize_po_file(trans_options, dir_path)

    # Save the number of nodes
    save_number_of_nodes(trans_options)

    # Save loop contexts
    save_loop_contexts(trans_options)

    # Save the edges in the partial order file
    po_file_path = trans_options.get_partial_order_file()
    edges = trans_options.get_all_edges()
    with open(po_file_path, "a") as po_file:
        for from_id, to_id in edges:
            po_file.write(serialize_edge(from_id, to_id))



from pash_annotations.datatypes.AccessKind import make_stream_output, make_stream_input
from pash_annotations.datatypes.BasicDatatypes import Flag, ArgStringType
from pash_annotations.datatypes.BasicDatatypesWithIO import OptionWithIO
from pash_annotations.datatypes.CommandInvocationWithIOVars import (
    CommandInvocationWithIOVars,
)

import os

import config
from ir_defs.arg import Arg
from ir_defs.dfg_node import DFGNode


class DGSHTee(DFGNode):
    def __init__(self, cmd_invocation_with_io_vars, com_redirs=None, com_assignments=None):
        com_redirs = [] if com_redirs is None else com_redirs
        com_assignments = [] if com_assignments is None else com_assignments
        super().__init__(
            cmd_invocation_with_io_vars,
            com_redirs=com_redirs,
            com_assignments=com_assignments,
        )


def make_dgsh_tee_node(input_id, output_id):
    dgsh_tee_bin = os.path.join(
        config.PASH_TOP, config.config["runtime"]["dgsh_tee_binary"]
    )

    access_map = {output_id: make_stream_output(), input_id: make_stream_input()}

    flag_option_list = [
        OptionWithIO("-i", input_id),
        OptionWithIO("-o", output_id),
        Flag("-I"),
        Flag("-f"),
        OptionWithIO(
            "-b",
            ArgStringType(
                Arg.string_to_arg(str(config.config["runtime"]["dgsh_buffer_size"]))
            ),
        ),
    ]

    cmd_inv_with_io_vars = CommandInvocationWithIOVars(
        cmd_name=dgsh_tee_bin,
        flag_option_list=flag_option_list,
        operand_list=[],
        implicit_use_of_streaming_input=None,
        implicit_use_of_streaming_output=None,
        access_map=access_map,
    )
    return DGSHTee(cmd_inv_with_io_vars)



"""
PaSh: Light-touch Data-Parallel Shell Processing

A system for parallelizing POSIX shell scripts.
Hosted by the Linux Foundation.

Usage:
    pash script.sh              # Run a script with PaSh
    pash -c "echo hi | cat"     # Run a command with PaSh
    pash --help                 # Show help

For more information, visit: https://github.com/binpash/pash
"""

__version__ = "0.13.0"
__author__ = "The PaSh Authors"
__license__ = "MIT"



import os

from pash_annotations.datatypes.AccessKind import (
    AccessKind,
    make_stream_input,
    make_stream_output,
)
from pash_annotations.datatypes.BasicDatatypes import Operand, Flag
from pash_annotations.datatypes.CommandInvocationWithIOVars import (
    CommandInvocationWithIOVars,
)

import config

from ir_defs.dfg_node import DFGNode
from ir_defs.arg import Arg


class RSplit(DFGNode):
    def __init__(
        self,
        cmd_invocation_with_io_vars,
        com_redirs=None,
        com_assignments=None,
        parallelizer_list=None,
        cmd_related_properties=None,
    ):
        com_redirs = [] if com_redirs is None else com_redirs
        com_assignments = [] if com_assignments is None else com_assignments
        super().__init__(
            cmd_invocation_with_io_vars=cmd_invocation_with_io_vars,
            com_redirs=com_redirs,
            com_assignments=com_assignments,
            parallelizer_list=parallelizer_list,
            cmd_related_properties=cmd_related_properties,
        )

    def add_r_flag(self):
        self.cmd_invocation_with_io_vars.flag_option_list.append(Flag("-r"))


def make_r_split(input_id, out_ids, r_split_batch_size):
    r_split_bin = os.path.join(
        config.PASH_TOP, config.config["runtime"]["r_split_binary"]
    )
    operand_list = [input_id, Operand(Arg.string_to_arg(str(r_split_batch_size)))]
    operand_list.extend(out_ids)
    access_map = {output_id: make_stream_output() for output_id in out_ids}
    access_map[input_id] = make_stream_input()
    cmd_inv_with_io_vars = CommandInvocationWithIOVars(
        cmd_name=r_split_bin,
        flag_option_list=[],
        operand_list=operand_list,
        implicit_use_of_streaming_input=None,
        implicit_use_of_streaming_output=None,
        access_map=access_map,
    )
    return RSplit(cmd_inv_with_io_vars)


def make_r_split_with_unwrap_flag(input_id, out_ids, r_split_batch_size):
    standard_r_split = make_r_split(input_id, out_ids, r_split_batch_size)
    standard_r_split.add_r_flag()
    return standard_r_split



"""
PaSh Compiler Package

Entry point:
- compilation_server.py: Compilation daemon/server (started by pa.sh)

Core API:
- compilation.compile_ir(): Main compilation function
- compilation.CompilerConfig: Compilation configuration

Argument Parsing:
- arg_parser.BaseParser: Base argument parser
- arg_parser.CompilerParser: Full compiler argument parser
"""

from .compilation import (
    compile_ir,
    CompilerConfig,
    compile_optimize_df_region,
    compile_optimize_output_script,
    load_df_region,
    optimize_irs,
)
from .ir import IR, FileIdGen

__all__ = [
    'compile_ir',
    'CompilerConfig',
    'compile_optimize_df_region',
    'compile_optimize_output_script',
    'load_df_region',
    'optimize_irs',
    'IR',
    'FileIdGen',
]



from typing import Set, List, Any

from ir_defs.arg import Arg

from pash_annotations.datatypes.CommandInvocationInitial import CommandInvocationInitial
from pash_annotations.datatypes.BasicDatatypes import (
    Option,
    ArgStringType,
    Flag,
    Operand,
)
from pash_annotations.parser.parser import (
    parse,
    get_set_of_all_flags,
    get_dict_flag_to_primary_repr,
    get_set_of_all_options,
    get_dict_option_to_primary_repr,
    are_all_individually_flags,
)
from pash_annotations.parser.util_parser import get_json_data


from util import format_arg_chars, string_to_argument


def merge_to_single_string_with_space(list_str):
    if len(list_str) == 1:
        return list_str[0]
    else:
        return " ".join(list_str)


def get_command_invocation(command, options) -> CommandInvocationInitial:
    command_as_string: str = format_arg_chars(command)
    options_and_operands_as_string: str = merge_to_single_string_with_space(
        [format_arg_chars(option) for option in options]
    )
    command_invocation_as_string: str = (
        f"{command_as_string} {options_and_operands_as_string}"
    )
    command_invocation: CommandInvocationInitial = parse(command_invocation_as_string)
    return command_invocation


def get_ast_for_flagoption(flagoption):
    result = string_to_argument(flagoption.get_name())
    if isinstance(flagoption, Option):
        # TODO: add argument here as well but eventually also fid
        assert False
    return result


def get_ast_for_argstringtype(arg):
    return string_to_argument(arg.get_name())


# TODO: this is a hack to fix the wrong parsing of "
def fix_parsing_newline(arg):
    if arg.get_name() == "\\n":
        return ArgStringType(r'"\n"')
    else:
        return arg


def parse_arg_list_to_command_invocation(
    command, flags_options_operands
) -> CommandInvocationInitial:
    cmd_name = format_arg_chars(command)
    json_data = get_json_data(cmd_name)

    set_of_all_flags: Set[str] = get_set_of_all_flags(json_data)
    dict_flag_to_primary_repr: dict[str, str] = get_dict_flag_to_primary_repr(json_data)
    set_of_all_options: Set[str] = get_set_of_all_options(json_data)
    dict_option_to_primary_repr: dict[str, str] = get_dict_option_to_primary_repr(
        json_data
    )
    # we keep the Arg for everything but flag and option names

    # parse list of command invocation terms
    flag_option_list: List[Any] = []
    i = 0
    while i < len(flags_options_operands):
        potential_flag_or_option_arg = flags_options_operands[i]
        potential_flag_or_option_name = format_arg_chars(potential_flag_or_option_arg)
        if potential_flag_or_option_name in set_of_all_flags:
            flag_name_as_string: str = dict_flag_to_primary_repr.get(
                potential_flag_or_option_name, potential_flag_or_option_name
            )
            flag: Flag = Flag(flag_name_as_string)
            flag_option_list.append(flag)
        elif (potential_flag_or_option_name in set_of_all_options) and (
            (i + 1) < len(flags_options_operands)
        ):
            option_name_as_string: str = dict_option_to_primary_repr.get(
                potential_flag_or_option_name, potential_flag_or_option_name
            )
            option_arg_as_arg: Arg = Arg(flags_options_operands[i + 1])
            option = Option(option_name_as_string, option_arg_as_arg)
            flag_option_list.append(option)
            i += 1  # since we consumed another term for the argument
        elif (
            potential_flag_or_option_name == "-"
        ):  # switch to operand mode (interpreted as hyphen-stdin)
            break
        elif are_all_individually_flags(
            potential_flag_or_option_name, set_of_all_flags
        ):
            for split_el in list(potential_flag_or_option_name[1:]):
                flag: Flag = Flag(f"-{split_el}")
                flag_option_list.append(flag)
        else:
            break  # next one is Operand, and we keep these in separate list
        i += 1

    # we would probably want to skip '--' but then the unparsed command could have a different meaning so we'd need to keep it
    # for now, omitted
    # if parsed_elements_list[i] == '--':
    #     i += 1

    operand_list = [
        Operand(Arg(operand_arg)) for operand_arg in flags_options_operands[i:]
    ]
    # log("type of operand_list[0].get_name()", type(operand_list[0].get_name()))   can only be used if there are operands

    return CommandInvocationInitial(cmd_name, flag_option_list, operand_list)



"""
PaSh JIT Runtime Module

This module provides utilities for finding JIT runtime shell scripts.

The JIT runtime handles:
- Just-in-time compilation of shell regions
- Shell state management during parallel execution
- Communication with the compilation server
"""

from pathlib import Path


def get_jit_runtime_dir() -> Path:
    """Get the JIT runtime directory containing shell scripts."""
    return Path(__file__).parent


def get_script_path(name: str) -> Path:
    """Get the path to a JIT runtime script.

    Args:
        name: Name of the script (e.g., 'jit.sh')

    Returns:
        Path to the script

    Raises:
        FileNotFoundError: If the script doesn't exist
    """
    script = get_jit_runtime_dir() / name
    if not script.exists():
        raise FileNotFoundError(
            f"JIT runtime script '{name}' not found at {script}."
        )
    return script


# Convenience functions for each script
def get_jit_script() -> Path:
    """Get path to the main jit.sh script."""
    return get_script_path("jit.sh")


def get_prepare_call_compiler_script() -> Path:
    """Get path to pash_prepare_call_compiler.sh."""
    return get_script_path("pash_prepare_call_compiler.sh")


def get_restore_state_script() -> Path:
    """Get path to pash_restore_state_and_execute.sh."""
    return get_script_path("pash_restore_state_and_execute.sh")


def get_set_from_to_script() -> Path:
    """Get path to pash_set_from_to.sh."""
    return get_script_path("pash_set_from_to.sh")


def get_source_declare_vars_script() -> Path:
    """Get path to pash_source_declare_vars.sh."""
    return get_script_path("pash_source_declare_vars.sh")



import subprocess
import os
import tempfile
import uuid


def read_file(file, mode="r"):
    with open(file, mode) as f:
        data = f.read()
    return data


def write_file(file, data, mode="w"):
    with open(file, mode) as f:
        n = f.write(data)
    return n


def create_filename(dir, prefix="", temp=False):
    if temp:
        return tempfile.mkstemp(dir=dir, prefix=prefix)
    else:
        return os.path.join(dir, f"{prefix}_{uuid.uuid4()}")



"""
Compilation daemon/server that handles multiple compilation requests.

This is the primary entry point when running PaSh via pa.sh.
"""

import logging
import os
import signal
import socket
from threading import Thread
from datetime import datetime, timedelta

# import queue

from sh_expand import env_vars_util

import config
from pash_graphviz import maybe_init_graphviz_dir, maybe_generate_graphviz
import ast_to_ir
import compilation
from dspash.worker_manager import WorkersManager

from arg_parser import BaseParser
from util import log, print_time_delta, NotAllRegionParallelizableError


##
## A Daemon (not with the strict Unix sense)
## that responds to requests for compilation
##

SOCKET_BUF_SIZE = 8192

SUCCESSFUL_COMPILATIONS = 0
TOTAL_REGIONS = 0
SUCCESSFUL_PARALLELIZATIONS = 0

def handler(signum, frame):
    log("Signal:", signum, "caught")
    shutdown()


signal.signal(signal.SIGTERM, handler)


##
## Functions that are used to communicate with the runtime
##

def success_response(string):
    return f"OK: {string}\n"


def error_response(string):
    return f"ERROR: {string}\n"


class UnixPipeReader:
    def __init__(self, in_filename, out_filename, blocking=True):
        self.in_filename = in_filename
        self.out_filename = out_filename
        self.buffer = ""
        self.blocking = blocking
        if not self.blocking:
            # Non blocking mode shouldn't be used in production. It's only used experimentally.
            log("Reader initialized in non-blocking mode")
            self.fin = open(self.in_filename)
        else:
            log("Reader initialized in blocking mode")

    ## This is necessary here to ensure that get_next_cmd is blocking een though the underlying API is non-blocking.
    def get_next_cmd(self):
        cmd = ""
        ## TODO: Remove the non-blocking control flow since it doesn't make sense to busy wait
        if not self.blocking:
            while not cmd:
                cmd = self.get_next_cmd_aux()
        else:
            cmd = self.get_next_cmd_aux()
        return cmd

    def get_next_cmd_aux(self):
        """
        This method return depends on the reading mode. In blocking mode this method will
        return the next full command and if there is no command it will wait until a full command is recieved.
        In non blocking mode it would either a full command or an empty string if a full command isn't available yet.
        This command keeps a state of the remaining data which is used in each subsequent call to this method.
        """
        input_buffer = ""
        if self.buffer:
            # Don't wait on fin if cmd buffer isn't empty
            log(
                "Reader buffer isn't empty. Using it instead of reading new data for the next command"
            )
            input_buffer = self.buffer
        else:
            log("Reader buffer is empty. Reading new data from input fifo")
            if self.blocking:
                with open(self.in_filename) as fin:
                    # This seems to be necessary for reading the full data.
                    # It seems like slower/smaller machines might not read the full data in one read
                    while True:
                        data = fin.read()
                        if len(data) == 0:
                            break
                        input_buffer += data
            else:
                input_buffer = self.fin.read()

        log("Input buffer:", input_buffer)
        if "\n" in input_buffer:
            cmd, rest = input_buffer.split("\n", 1)  # split on the first \n only
            self.buffer = rest
        else:
            cmd = input_buffer
            self.buffer = ""

        cmd = cmd.rstrip()
        log("Reader returned cmd:", cmd)
        return cmd

    ## This method respond to the connection we last got input from
    ## In the case of the UnixPipes, we don't have any state management here
    ##   since all reads/writes go to/from the same fifos
    def respond(self, message):
        fout = open(self.out_filename, "w")
        fout.write(message)
        fout.flush()
        fout.close()

    ## This method doesn't do anything for unix pipe reader since we always read and write
    ## to and from the same fifos
    def close_last_connection(self):
        pass

    def close(self):
        log("Reader closed")
        if not self.blocking:
            self.fin.close()


def unix_socket_send_and_forget(socket_file: str, msg: str):
    try:
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        sock.connect(socket_file)
        msg_with_newline = msg + "\n"
        byte_msg = msg_with_newline.encode("utf-8")
        sock.sendall(byte_msg)
        data = sock.recv(SOCKET_BUF_SIZE)
        str_data = data.decode("utf-8")
        ## There should be no response on these messages
        assert len(str_data) == 0
    finally:
        log("Sent message:", msg, "to server.")
        sock.close()


## TODO: Instead of this, think of using a standard SocketServer
##   see: https://docs.python.org/3/library/socketserver.html#module-socketserver
##
## TODO: SocketManager might need to handle errors more gracefully
class SocketManager:
    def __init__(self, socket_addr: str):
        ## Configure them outside
        server_address = socket_addr
        self.buf_size = SOCKET_BUF_SIZE

        # Make sure the socket does not already exist
        ## TODO: Is this necessary?
        try:
            os.unlink(server_address)
        except OSError:
            if os.path.exists(server_address):
                raise
        log("SocketManager: Made sure that socket does not exist")

        # Create a UDS socket
        self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        log("SocketManager: Created socket")

        self.sock.bind(server_address)
        log("SocketManager: Successfully bound to socket")

        ## TODO: Check if we need to configure the backlog
        self.sock.listen()
        log("SocketManager: Listenting on socket")

        ## Connection stack
        self.connections = []

    def get_next_cmd(self):
        connection, client_address = self.sock.accept()
        data = connection.recv(self.buf_size)

        ## TODO: This could be avoided for efficiency
        str_data = data.decode("utf-8")
        log("Received data:", str_data)
        ## TODO: Lift this requirement if needed
        ##
        ## We need to ensure that we read a command at once or the command was empty (only relevant in the first invocation)
        assert str_data.endswith("\n") or str_data == ""

        self.connections.append(connection)
        return str_data

    ## This method respond to the connection we last got input from
    ## In the case of the UnixPipes, we don't have any state management here
    ##   since all reads/writes go to/from the same fifos
    def respond(self, message):
        bytes_message = message.encode("utf-8")
        self.connections[-1].sendall(bytes_message)
        self.close_last_connection()

    ## This method doesn't do anything for unix pipe reader since we always read and write
    ## to and from the same fifos
    def close_last_connection(self):
        # Clean up the connection
        last_connection = self.connections.pop()
        last_connection.close()

    def close(self):
        self.sock.close()
        log("SocketManager: Closed")



def parse_args():
    parser = BaseParser(add_help=False)
    parser.add_pash_args()
    args, _ = parser.parse_known_args()

    return args


# Initialize the daemon


def init():
    ## Set the logging prefix
    config.LOGGING_PREFIX = "Daemon: "

    args = parse_args()
    config.set_config_globals_from_pash_args(args)

    # Load the configuration
    if not config.config:
        config.load_config(args.config_path)

    compilation.runtime_config = config.config["distr_planner"]

    ## Initialize the graphviz directory
    maybe_init_graphviz_dir(args)

    return args


##
## This class holds information for each process id
##
class ProcIdInfo:
    def __init__(self, input_ir, compiler_config, exec_time=None, start_exec_time=None):
        self.input_ir = input_ir
        self.compiler_config = compiler_config
        self.exec_time = exec_time
        self.start_exec_time = start_exec_time
        ## TODO: Extend it with other info from scheduler, like dependencies

    def set_exec_time(self, exec_time):
        self.exec_time = exec_time

    def set_start_exec_time(self, start_exec_time):
        self.start_exec_time = start_exec_time

    def get_start_exec_time(self):
        return self.start_exec_time

    def __repr__(self):
        return f"ProcIdInfo(InputIR:{self.input_ir}, CompConfig:{self.compiler_config}, ExecTime:{self.exec_time})"


class Scheduler:
    """Takes care of running processes in parallel if there is no conflict.
    The scheduler relies on the fact that process will wait for a compilation response.
    This allows it to control wether to allow the next process to run or wait for all other process.
    Flow:
        input cmd ->
                    |   Compile ->
                            1- Try compiling the pipeline
                            2- Wait for any unsafe processes to finish
                            3- Check compilation for success and any conficts
                                - no side effects -> allow to run in parallel by sending a response
                                - failed or conflict -> wait for all process to exit then run this process in unsafe mode

                    |   Exit process_id -> remove process_id from the list of running processes

                    |   Done -> no more pipelines -> wait for all processes to finish and exit

    Notes:
        The current design relies on the data being written atomicly to the pipe for the size of our input. This allows as to use only one pipe instead of a pipe per process.
        Source: https://www.gnu.org/software/libc/manual/html_node/Pipe-Atomicity.html
    """

    def __init__(self):
        self.input_resources = set()
        self.output_resources = set()
        self.process_resources = (
            {}
        )  # map process_id -> (input_resources, output_resources)
        self.next_id = 0
        self.running_procs = 0
        self.unsafe_running = False
        self.done = False
        self.cmd_buffer = ""
        self.connection_manager = None
        self.reader_pipes_are_blocking = True
        self.request_processing_start_time = 0
        ## TODO: Make that be a class or something

        ## A map that keeps mappings between proc_id and (input_ir, width, exec_time)
        self.process_id_input_ir_map = {}
        ## This is a map from input IRs, i.e., locations in the code, to a list of process_ids
        self.input_ir_to_process_id_map = {}

    def check_resources_safety(self, process_id):
        proc_input_resources, proc_output_resources = self.process_resources[process_id]
        all_proc_resources = proc_input_resources.union(proc_output_resources)
        if self.output_resources.intersection(
            all_proc_resources
        ) or self.input_resources.intersection(proc_output_resources):
            return False
        return True

    ##############################################################################
    ##
    ## Profiling-based Compiler-config
    ##
    ##############################################################################

    ## TODO: It is possible that there will be many process ids being gathered here.
    ##       If that becomes an overhead, we need to consider removing old ones
    ##         or keeping sketches of the information (instead of their raw numbers).

    def determine_compiler_config(self, input_ir_file):
        if config.pash_args.profile_driven:
            ## A default value
            selected_width = config.pash_args.width
            # log("Using profiles to select width!")

            ## Goal: Find the highest width that gives benefits
            ##
            ## Strategy, start trying lower widths, if the time seems to drop, keep trying lower.
            ##
            width_avgs = self.get_averages_per_width(input_ir_file)
            log("Width averages:", width_avgs)
            widths = width_avgs.keys()

            ## If we have at least 1, with a specific width,
            ##   and the minimum width has the lowest average, then try one lower
            if len(widths) > 0:
                min_width = min(widths)
                min_width_average = width_avgs[min_width]
                # log("Minimum width:", min_width, "with average:", min_width_average)

                best_width = min_width
                best_width_average = min_width_average
                for width, average in width_avgs.items():
                    ## We find the width with the best average
                    ## If that ends up being the minimum, then we also try one lower.
                    if average < best_width_average:
                        best_width = width
                        best_width_average = average

                if best_width == min_width and min_width > 1:
                    ## Divide the min_width by 2 and try again
                    selected_width = min_width // 2
                    log(
                        "Best width is the lowest width, trying with width:",
                        selected_width,
                    )
                else:
                    selected_width = best_width
                    log("Best width is:", best_width, "We will keep executing with it.")
        else:
            selected_width = config.pash_args.width

        log("Selected width:", selected_width)
        return compilation.CompilerConfig(selected_width)

    def get_averages_per_width(self, input_ir_file):
        ## If we haven't gathered any statistic yet
        if not input_ir_file in self.input_ir_to_process_id_map:
            return {}

        all_proc_ids = self.input_ir_to_process_id_map[input_ir_file]

        ## TODO: If that becomes an overhead, we need to add a new
        ##       data structure to keep the latest averages (sum, len)
        ##       to avoid recomputing every time.
        width_times = {}
        for proc_id in all_proc_ids:
            proc_info = self.process_id_input_ir_map[proc_id]
            width = proc_info.compiler_config.width
            exec_time = proc_info.exec_time

            if exec_time is not None:
                try:
                    width_times[width].append(exec_time)
                except:
                    width_times[width] = [exec_time]

        ## We have gathered all times for each width
        width_avgs = {}
        for width, exec_times in width_times.items():
            width_avgs[width] = sum(exec_times) / len(exec_times)

        return width_avgs

    ## This adds the time measurement, or just removes the entry if there is no exec_time (for space reclamation)
    def handle_time_measurement(self, process_id, exec_time):
        ## 2023-12-08 KK: When in parallel pipelines we receive two exits (when I tried to make it one something got stuck...)
        ##                so this assert is not true
        # assert self.process_id_input_ir_map[process_id].exec_time is None

        ## If we don't have the exec time we do Nothing
        ##
        ## TODO: Consider removing past entries that have no execution time.
        if exec_time is None:
            pass
        else:
            self.process_id_input_ir_map[process_id].set_exec_time(exec_time)

        # log("All measurements:", self.process_id_input_ir_map)

    def add_proc_id_map(self, process_id, input_ir_file, compiler_config):
        assert not process_id in self.process_id_input_ir_map
        self.process_id_input_ir_map[process_id] = ProcIdInfo(
            input_ir_file, compiler_config
        )

        ## Add the mapping from ir to process_id
        try:
            self.input_ir_to_process_id_map[input_ir_file].append(process_id)
        except:
            self.input_ir_to_process_id_map[input_ir_file] = [process_id]
        # log("All mappings from input_ir to proc_id:", self.input_ir_to_process_id_map)

    ##############################################################################

    def compile_and_add(self, compiled_script_file, var_file, input_ir_file):
        global TOTAL_REGIONS, SUCCESSFUL_PARALLELIZATIONS
        TOTAL_REGIONS += 1

        process_id = self.get_next_id()
        run_parallel = False
        compile_success = False
        current_region_parallelizable = True

        variable_reading_start_time = datetime.now()
        # Read any shell variables files if present
        vars_dict = env_vars_util.read_vars_file(var_file, config.BASH_VERSION)
        config.set_vars_file(var_file, vars_dict)

        variable_reading_end_time = datetime.now()
        print_time_delta(
            "Variable Loading", variable_reading_start_time, variable_reading_end_time
        )

        daemon_compile_start_time = datetime.now()
        ## TODO: Make the compiler config based on profiling data
        compiler_config = self.determine_compiler_config(input_ir_file)
        ## Add the process_id -> input_ir mapping
        self.add_proc_id_map(process_id, input_ir_file, compiler_config)

        # check if any general exceptions are caught to report to --assert_compiler_success flag
        try:
            ast_or_ir = compilation.compile_ir(
                input_ir_file, compiled_script_file, config.pash_args, compiler_config
            )
        except NotAllRegionParallelizableError:
            ast_or_ir = None
            current_region_parallelizable = False

        # Track successful parallelizations (region was parallelizable)
        if current_region_parallelizable:
            SUCCESSFUL_PARALLELIZATIONS += 1

        daemon_compile_end_time = datetime.now()
        print_time_delta(
            "Daemon Compile", daemon_compile_start_time, daemon_compile_end_time
        )

        self.wait_unsafe()
        if ast_or_ir != None:
            compile_success = True

            ## Increase the counter of successful compilations
            global SUCCESSFUL_COMPILATIONS
            SUCCESSFUL_COMPILATIONS += 1

            maybe_generate_graphviz(
                ast_or_ir, config.pash_args, name=f"dfg-{process_id}"
            )

            proc_input_resources = set(
                map(
                    lambda out: str(out.resource)
                    if str(out.resource) != "None"
                    else out,
                    ast_or_ir.all_input_fids(),
                )
            )
            proc_output_resources = set(
                map(
                    lambda out: str(out.resource)
                    if str(out.resource) != "None"
                    else out,
                    ast_or_ir.all_output_fids(),
                )
            )

            self.process_resources[process_id] = (
                proc_input_resources,
                proc_output_resources,
            )

            run_parallel = self.check_resources_safety(process_id)
            if run_parallel:
                self.input_resources = self.input_resources.union(proc_input_resources)
                self.output_resources = self.output_resources.union(
                    proc_output_resources
                )

        if not run_parallel:
            ## If we are not running in parallel everything has to finish first before scheduling for execution
            self.wait_for_all()
        else:
            ## Wait if we have more pipelines running than our current limit
            self.wait_until_limit(config.pash_args.parallel_pipelines_limit)

        if compile_success:
            response = success_response(
                f"{process_id} {compiled_script_file} {var_file} {input_ir_file}"
            )
        elif not current_region_parallelizable:
            # send specified message to say current region is not parallelizable instead of general exception caught
            response = error_response(f"{process_id} current region is not parallelizable; failed to compile")
            self.unsafe_running = True
        else:
            response = error_response(f"{process_id} failed to compile")
            self.unsafe_running = True

        self.running_procs += 1

        ## Get the time before we start executing (roughly) to determine how much time this command execution will take
        command_exec_start_time = datetime.now()
        self.process_id_input_ir_map[process_id].set_start_exec_time(
            command_exec_start_time
        )
        return response

    def remove_process(self, process_id):
        log("The following process exited:", process_id)
        if process_id in self.process_resources:
            del self.process_resources[process_id]
            # TODO: Should be improved to not rebuild inputs and outputs from scratch maybe use counters
            self.input_resources = set().union(
                *[
                    input_resources
                    for input_resources, _ in self.process_resources.values()
                ]
            )
            self.output_resources = set().union(
                *[
                    output_resources
                    for _, output_resources in self.process_resources.values()
                ]
            )

        assert self.running_procs > 0
        self.running_procs -= 1
        if self.running_procs == 0:
            self.unsafe_running = False

    def get_next_id(self):
        self.next_id += 1
        return self.next_id

    def wait_for_all(self):
        log(
            "Waiting for all processes to finish."
        )
        self.wait_until_limit(1)
        self.unsafe_running = False

    def wait_until_limit(self, limit: int):
        log(
            f"Waiting for less than {limit} processes to be running. There are",
            self.running_procs,
            "processes remaining.",
        )
        while self.running_procs >= limit:
            input_cmd = self.get_input()
            # must be exit command or something is wrong
            if input_cmd.startswith("Exit:"):
                self.handle_exit(input_cmd)
            else:
                raise Exception(f"Command should be exit but it was {input_cmd}")

    def handle_exit(self, input_cmd):
        assert input_cmd.startswith("Exit:")
        process_id = int(input_cmd.split(":")[1])

        ## Get the execution time
        command_finish_exec_time = datetime.now()
        command_start_exec_time = self.process_id_input_ir_map[
            process_id
        ].get_start_exec_time()
        exec_time = (command_finish_exec_time - command_start_exec_time) / timedelta(
            milliseconds=1
        )
        log("Process:", process_id, "exited. Exec time was:", exec_time)
        self.handle_time_measurement(process_id, exec_time)
        self.remove_process(process_id)
        ## Necessary so that Exit doesn't block
        self.close_last_connection()

    def wait_unsafe(self):
        log("Unsafe running:", self.unsafe_running)
        if self.unsafe_running:
            assert self.running_procs == 1
            self.wait_for_all()
            self.unsafe_running = False

    def parse_and_run_cmd(self, input_cmd):
        if input_cmd.startswith("Compile"):
            (
                compiled_script_file,
                var_file,
                input_ir_file,
            ) = self.__parse_compile_command(input_cmd)
            response = self.compile_and_add(
                compiled_script_file, var_file, input_ir_file
            )
            request_processing_end_time = datetime.now()
            print_time_delta(
                "Request handling",
                self.request_processing_start_time,
                request_processing_end_time,
            )
            ## Send output to the specific command
            self.respond(response)
        elif input_cmd.startswith("Exit:"):
            self.handle_exit(input_cmd)
        elif input_cmd.startswith("Done"):
            self.wait_for_all()
            ## Check if any assertion failed
            assertion_failed = False
            if config.pash_args.assert_all_regions_parallelizable and SUCCESSFUL_PARALLELIZATIONS != TOTAL_REGIONS:
                assertion_failed = True
            if config.pash_args.assert_compiler_success and SUCCESSFUL_COMPILATIONS != TOTAL_REGIONS:
                assertion_failed = True
            ## We send output to the top level pash process
            ## to signify that we are done.
            if assertion_failed:
                self.respond("ASSERT_FAILED")
            else:
                self.respond("All finished")
            self.done = True
        elif input_cmd.startswith("Daemon Start") or input_cmd == "":
            ## This happens when pa.sh first connects to daemon to see if it is on
            self.close_last_connection()
        else:
            log(error_response(f"Error: Unsupported command: {input_cmd}"))
            raise Exception(f"Error: Unsupported command: {input_cmd}")

    ## This method calls the reader to get an input
    def get_input(self):
        return self.connection_manager.get_next_cmd()

    ## This method responds to the last connection and closes it
    def respond(self, message):
        self.connection_manager.respond(message)

    ## This method closes the last connection that we got input from
    def close_last_connection(self):
        self.connection_manager.close_last_connection()

    def __parse_compile_command(self, input):
        try:
            components = input.rstrip().split("|")
            compiled_script_file = components[0].split(":")[1]
            var_file = components[1].split(":")[1]
            input_ir_file = components[2].split(":")[1]
            return compiled_script_file, var_file, input_ir_file
        except:
            raise Exception(f"Parsing failure for line: {input}")

    def run(self):
        ## By default communicate through sockets, except if the user wants to do it through pipes
        if config.pash_args.daemon_communicates_through_unix_pipes:
            in_filename = os.getenv("RUNTIME_IN_FIFO")
            out_filename = os.getenv("RUNTIME_OUT_FIFO")
            self.connection_manager = UnixPipeReader(
                in_filename, out_filename, self.reader_pipes_are_blocking
            )
        else:
            self.connection_manager = SocketManager(
                os.getenv("DAEMON_SOCKET")
            )
        while not self.done:
            # Process a single request
            input_cmd = self.get_input()
            self.request_processing_start_time = datetime.now()

            ## Parse the command (potentially also sending a response)
            self.parse_and_run_cmd(input_cmd)

        self.connection_manager.close()
        shutdown()


def shutdown():
    global SUCCESSFUL_COMPILATIONS, TOTAL_REGIONS, SUCCESSFUL_PARALLELIZATIONS

    # in-bash expansion server, if it exists
    try:
        ast_to_ir.BASH_EXP_STATE
    except AttributeError:
        pass
    else:
        ast_to_ir.BASH_EXP_STATE.close()

    # Report assertion failures
    if config.pash_args.assert_all_regions_parallelizable:
        if SUCCESSFUL_PARALLELIZATIONS != TOTAL_REGIONS:
            logging.warning(f"[PaSh] Assertion failed: Not all regions were parallelizable ({SUCCESSFUL_PARALLELIZATIONS}/{TOTAL_REGIONS} parallelized)")

    if config.pash_args.assert_compiler_success:
        if SUCCESSFUL_COMPILATIONS != TOTAL_REGIONS:
            logging.warning(f"[PaSh] Assertion failed: Not all regions were compiled successfully ({SUCCESSFUL_COMPILATIONS}/{TOTAL_REGIONS} compiled)")

    # Default warning when no assertions and no successful compilations
    if not config.pash_args.assert_all_regions_parallelizable and not config.pash_args.assert_compiler_success:
        if SUCCESSFUL_COMPILATIONS == 0 and TOTAL_REGIONS > 0:
            logging.warning("[PaSh] Warning: No region was parallelized successfully")
            logging.warning("       Hint: Check if there are annotations for the commands in your script")

    ## There may be races since this is called through the signal handling
    log("PaSh daemon is shutting down...")
    log("PaSh daemon shut down successfully...")


def main():
    args = init()
    if args.distributed_exec:
        worker_manager = WorkersManager()
        config.worker_manager = worker_manager
        worker_manager_thread = Thread(target=worker_manager.run)
        worker_manager_thread.start()

    scheduler = Scheduler()
    scheduler.run()


if __name__ == "__main__":
    main()



"""
Utility functions for the PaSh preprocessor.

This is a simplified version of compiler/util.py that doesn't depend on config.py.
Configuration values are read from environment variables.
"""

from datetime import timedelta
import functools
import logging
import os
import tempfile

# Configuration from environment variables (set by pa.sh or pash_runtime.sh)
PASH_TMP_PREFIX = os.environ.get("PASH_TMP_PREFIX", "/tmp/pash_tmp/")
OUTPUT_TIME = os.environ.get("pash_output_time_flag", "1") == "1"
LOGGING_PREFIX = "PaSh: "


def unzip(lst):
    """Unzip a list of pairs into two separate lists."""
    res = [[i for i, j in lst], [j for i, j in lst]]
    return res


def print_time_delta(prefix, start_time, end_time):
    """Output timing information to the log."""
    time_difference = (end_time - start_time) / timedelta(milliseconds=1)
    if OUTPUT_TIME:
        log("{} time:".format(prefix), time_difference, " ms", level=0)
    else:
        log("{} time:".format(prefix), time_difference, " ms")


def logging_prefix(logging_prefix_str):
    """Decorator to add logging prefix to a function."""
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            global LOGGING_PREFIX
            old_prefix = LOGGING_PREFIX
            LOGGING_PREFIX = logging_prefix_str
            result = func(*args, **kwargs)
            LOGGING_PREFIX = old_prefix
            return result
        return wrapper
    return decorator


def log(*args, end="\n", level=2):
    """Wrapper for logging."""
    if level == 1:
        concatted_args = " ".join([str(a) for a in list(args)])
        logging.warning(f"{LOGGING_PREFIX} {concatted_args}")
    elif level >= 2:
        concatted_args = " ".join([str(a) for a in list(args)])
        logging.info(f"{LOGGING_PREFIX} {concatted_args}")


def ptempfile():
    """Create a temporary file in the PaSh temp directory."""
    fd, name = tempfile.mkstemp(dir=PASH_TMP_PREFIX)
    os.close(fd)
    return name


def make_kv(key, val):
    """Make a key-value pair in AST JSON format."""
    return [key, val]



from pash_annotations.datatypes.CommandInvocationWithIOVars import (
    CommandInvocationWithIOVars,
)
from ir_defs.dfg_node import DFGNode


def make_cat_node(inputs, output):
    cmd_inv_cat = CommandInvocationWithIOVars.make_cat_command_invocation_with_io_vars(
        inputs, output
    )
    return DFGNode.make_simple_dfg_node_from_cmd_inv_with_io_vars(cmd_inv_cat)



from typing import List

import pash_annotations.datatypes

from pash_annotations.datatypes.CommandInvocationInitial import CommandInvocationInitial
from pash_annotations.datatypes.BasicDatatypes import ArgStringType
from pash_annotations.datatypes.BasicDatatypesWithIO import (
    FileNameWithIOInfo,
    StdDescriptorWithIOInfo,
    OptionWithIO,
)
from pash_annotations.annotation_generation.datatypes.InputOutputInfo import (
    InputOutputInfo,
)
from pash_annotations.annotation_generation.datatypes.ParallelizabilityInfo import (
    ParallelizabilityInfo,
)
from pash_annotations.annotation_generation.datatypes.CommandProperties import (
    CommandProperties,
)
from pash_annotations.datatypes.CommandInvocationWithIOVars import (
    CommandInvocationWithIOVars,
)

from shasta.ast_node import ast_node_to_untyped_deep
from shasta.json_to_ast import to_ast_node

from annotations_utils.util_parsing import parse_arg_list_to_command_invocation
from annotations_utils.util_cmd_invocations import (
    get_input_output_info_from_cmd_invocation_util,
    get_parallelizability_info_from_cmd_invocation_util,
)
from annotations_utils.util_file_descriptors import resource_from_file_descriptor

from ir_defs.file_id import *
from ir_defs.nodes.cat import *

import ir_defs.nodes.pash_split as pash_split
import ir_defs.nodes.r_merge as r_merge
import ir_defs.nodes.r_split as r_split
import ir_defs.nodes.r_wrap as r_wrap
import ir_defs.nodes.r_unwrap as r_unwrap

from util import (
    log,
    flatten_list,
    UnparallelizableError,
    AdjLineNotImplementedError,
    format_arg_chars,
    string_to_argument,
    char_to_arg_char,
    standard_var_ast,
    quote_arg,
    make_background,
    make_command,
    make_assignment,
)

import config


## Creates a file id for a given resource
def create_file_id_for_resource(resource, fileIdGen):
    file_id = create_split_file_id(fileIdGen)
    file_id.set_resource(resource)
    return file_id


## Creates a file id that has a given maximum length
def create_split_file_id(fileIdGen):
    file_id = fileIdGen.next_file_id()
    return file_id


class FileIdGen:
    def __init__(self, next=0, prefix=""):
        self.next = next + 1
        directory = f"{str(uuid.uuid4().hex)}"
        self.prefix = f"{directory}/{prefix}"
        directory_path = os.path.join(config.PASH_TMP_PREFIX, self.prefix)
        os.makedirs(directory_path)

    def next_file_id(self):
        fileId = FileId(self.next, self.prefix)
        self.next += 1
        return fileId

    def next_temporary_file_id(self):
        fileId = self.next_file_id()
        fileId.make_temporary_file()
        return fileId

    def next_ephemeral_file_id(self):
        fileId = self.next_file_id()
        fileId.make_ephemeral()
        return fileId

    def bump_counter_to_value_of(self, OtherFileIdGen):
        # TODO: find a better solution to make unique numbers, currently: set to max-value + 1
        self.next = OtherFileIdGen.next + 1


## Returns the resource or file descriptor related to this specific opt_or_fd
## NOTE: Assumes that everything is expanded.
def get_option_or_fd(opt_or_fd, options, fileIdGen):
    if (
        isinstance(opt_or_fd, tuple)
        and len(opt_or_fd) == 2
        and opt_or_fd[0] == "option"
    ):
        resource = FileResource(Arg(options[opt_or_fd[1]]))
    else:
        ## TODO: Make this be a subtype of Resource
        if opt_or_fd == "stdin":
            resource = ("fd", 0)
        elif opt_or_fd == "stdout":
            resource = ("fd", 1)
        elif opt_or_fd == "stderr":
            resource = ("fd", 2)
        else:
            raise NotImplementedError()
        resource = FileDescriptorResource(resource)

    fid = create_file_id_for_resource(resource, fileIdGen)
    return fid


## Get the options as arguments
def get_option(opt_or_fd, options, fileIdGen):
    assert (
        isinstance(opt_or_fd, tuple)
        and len(opt_or_fd) == 2
        and opt_or_fd[0] == "option"
    )
    arg = Arg(options[opt_or_fd[1]])
    return (opt_or_fd[1], arg)


## This function
def create_edges_from_opt_or_fd_list(opt_or_fd_list, edges_dict, options, fileIdGen):
    new_edge_list = []
    for opt_or_fd in opt_or_fd_list:
        fid = get_option_or_fd(opt_or_fd, options, fileIdGen)
        fid_id = fid.get_ident()
        edges_dict[fid_id] = (fid, None, None)
        new_edge_list.append(fid_id)
    return new_edge_list


def find_input_edges(
    positional_input_list, implicit_use_of_stdin, dfg_edges, fileIdGen
) -> List[int]:
    assert not implicit_use_of_stdin or len(positional_input_list) == 0
    if implicit_use_of_stdin:
        resources = [FileDescriptorResource(("fd", 0))]
    else:
        resources = [
            resource_from_file_descriptor(input_el)
            for input_el in positional_input_list
        ]
    file_ids = [
        create_file_id_for_resource(resource, fileIdGen) for resource in resources
    ]
    return get_edge_list_from_file_id_list(dfg_edges, file_ids)


def find_output_edges(
    positional_output_list, implicit_use_of_stdout, dfg_edges, fileIdGen
) -> List[int]:
    assert not implicit_use_of_stdout or len(positional_output_list) == 0
    if implicit_use_of_stdout:
        resources = [FileDescriptorResource(("fd", 1))]
    else:
        resources = [
            resource_from_file_descriptor(input_el)
            for input_el in positional_output_list
        ]
    file_ids = [
        create_file_id_for_resource(resource, fileIdGen) for resource in resources
    ]
    return get_edge_list_from_file_id_list(dfg_edges, file_ids)


def get_edge_list_from_file_id_list(dfg_edges, file_ids):
    new_edge_list = []
    for file_id in file_ids:
        fid_id = file_id.get_ident()
        dfg_edges[fid_id] = (file_id, None, None)
        new_edge_list.append(fid_id)
    return new_edge_list


def add_file_id_vars(command_invocation_with_io, fileIdGen):
    # make pass over everything and create file_id for everything
    # only for operands for now:
    dfg_edges = {}
    new_flagoption_list = []
    new_operand_list = []
    access_map = dict()

    def add_var_for_descriptor(operand):
        resource = resource_from_file_descriptor(operand)
        file_id = create_file_id_for_resource(resource, fileIdGen)
        fid_id = file_id.get_ident()
        dfg_edges[fid_id] = (file_id, None, None)
        access_map[fid_id] = operand.get_access()
        return fid_id

    for i in range(len(command_invocation_with_io.flag_option_list)):
        flagoption = command_invocation_with_io.flag_option_list[i]
        if isinstance(flagoption, OptionWithIO) and not isinstance(
            flagoption.option_arg, ArgStringType
        ):
            fid_id = add_var_for_descriptor(flagoption.option_arg)
            new_option = OptionWithIOVar(flagoption.name, fid_id)
            new_flagoption_list.append(new_option)
        else:  # Flag
            new_flagoption_list.append(flagoption)

    for i in range(len(command_invocation_with_io.operand_list)):
        operand = command_invocation_with_io.operand_list[i]
        if isinstance(operand, FileNameWithIOInfo) or isinstance(
            operand, StdDescriptorWithIOInfo
        ):
            fid_id = add_var_for_descriptor(operand)
            new_operand_list.append(fid_id)
        else:
            new_operand_list.append(operand)
    if command_invocation_with_io.implicit_use_of_streaming_input:
        new_implicit_use_of_streaming_input = add_var_for_descriptor(
            command_invocation_with_io.implicit_use_of_streaming_input
        )
    else:
        new_implicit_use_of_streaming_input = None
    if command_invocation_with_io.implicit_use_of_streaming_output:
        new_implicit_use_of_streaming_output = add_var_for_descriptor(
            command_invocation_with_io.implicit_use_of_streaming_output
        )
    else:
        new_implicit_use_of_streaming_output = None

    command_invocation_with_io_vars = CommandInvocationWithIOVars(
        cmd_name=command_invocation_with_io.cmd_name,
        flag_option_list=new_flagoption_list,
        operand_list=new_operand_list,
        implicit_use_of_streaming_input=new_implicit_use_of_streaming_input,
        implicit_use_of_streaming_output=new_implicit_use_of_streaming_output,
        access_map=access_map,
    )
    return command_invocation_with_io_vars, dfg_edges


def compile_command_to_DFG(fileIdGen, command, options, redirections=None):
    redirections = [] if redirections is None else redirections
    command_invocation: CommandInvocationInitial = parse_arg_list_to_command_invocation(
        command, options
    )
    io_info: InputOutputInfo = get_input_output_info_from_cmd_invocation_util(
        command_invocation
    )
    if io_info is None:
        raise UnparallelizableError(
            f"InputOutputInformation for {format_arg_chars(command)} not provided so considered side-effectful."
        )
    if io_info.has_other_outputs():
        raise UnparallelizableError(
            f"Command {format_arg_chars(command)} has outputs other than streaming."
        )
    para_info: ParallelizabilityInfo = (
        get_parallelizability_info_from_cmd_invocation_util(command_invocation)
    )
    if para_info is None:
        para_info = (
            ParallelizabilityInfo()
        )  # defaults to no parallelizer's and all properties False
    command_invocation_with_io = io_info.apply_input_output_info_to_command_invocation(
        command_invocation
    )
    if para_info is None:
        para_info = (
            ParallelizabilityInfo()
        )  # defaults to no parallelizer's and all properties False
    (
        parallelizer_list,
        round_robin_compatible_with_cat,
        is_commutative,
    ) = para_info.unpack_info()
    property_dict = [
        {
            "round_robin_compatible_with_cat": round_robin_compatible_with_cat,
            "is_commutative": is_commutative,
        }
    ]
    cmd_related_properties = CommandProperties(property_dict)

    ## TODO: Make an empty IR and add edges and nodes incrementally (using the methods defined in IR).

    ## Add all inputs and outputs to the DFG edges
    cmd_invocation_with_io_vars, dfg_edges = add_file_id_vars(
        command_invocation_with_io, fileIdGen
    )
    com_redirs = redirections
    ## TODO: Add assignments
    com_assignments = []

    ## Assume: Everything must be completely expanded
    ## TODO: Add an assertion about that.
    dfg_node = DFGNode(
        cmd_invocation_with_io_vars,
        com_redirs=com_redirs,
        com_assignments=com_assignments,
        parallelizer_list=parallelizer_list,
        cmd_related_properties=cmd_related_properties,
    )
    # log(f'Dfg node: {dfg_node}')
    node_id = dfg_node.get_id()

    ## Assign the from, to node in edges
    for fid_id in dfg_node.get_input_list():
        fid, from_node, to_node = dfg_edges[fid_id]
        assert to_node is None
        dfg_edges[fid_id] = (fid, from_node, node_id)

    for fid_id in dfg_node.get_output_list():
        fid, from_node, to_node = dfg_edges[fid_id]
        assert from_node is None
        dfg_edges[fid_id] = (fid, node_id, to_node)

    dfg_nodes = {node_id: dfg_node}
    dfg = IR(dfg_nodes, dfg_edges)
    # log(f'IR: {dfg}')
    return dfg


##
## Node builder functions
##


def make_tee(input, outputs):
    com_name = Arg.string_to_arg("tee")
    com_category = "pure"
    return DFGNode([input], outputs, com_name, com_category)


## Note: This might need more information. E.g. all the file
## descriptors of the IR, and in general any other local information
## that might be relevant.
class IR:
    ## TODO: Embed the fileIdGen as a field of the IR

    ## IR Assumptions:
    ##
    ## - Each node has a list of incoming files in order of
    ##   consumption.
    ##
    ## - If two nodes have the same file as output, then they both
    ##   write to it concurrently.
    def __init__(self, nodes, edges, background=False):
        self.nodes = nodes
        self.edges = edges
        self.background = background

        ## Apply the redirections for each separate node.
        ## This needs to be called here because nodes do not
        ## have information about the edges on their own.
        self.apply_redirections()

    def __repr__(self):
        output = "(|-{} IR: {} {}-|)".format(
            self.get_stdin(), list(self.nodes.values()), self.get_stdout()
        )
        return output

    ## Initialize all edges
    def apply_redirections(self):
        for _, node in self.nodes.items():
            node.apply_redirections(self.edges)

        ## We need to merge common files after redirections have been applied.
        self.combine_common_files()

    ## Refactor these to call .add_edge, and .set_edge_to/from
    ## Add an edge that points to a node
    def add_to_edge(self, to_edge, node_id):
        edge_id = to_edge.get_ident()
        assert not edge_id in self.edges
        self.edges[edge_id] = (to_edge, None, node_id)

    ## Add an edge that starts from a node
    def add_from_edge(self, node_id, from_edge):
        edge_id = from_edge.get_ident()
        assert not edge_id in self.edges
        self.edges[edge_id] = (from_edge, node_id, None)

    def set_edge_to(self, edge_id, to_node_id):
        edge_fid, from_node, old_to_node = self.edges[edge_id]
        self.edges[edge_id] = (edge_fid, from_node, to_node_id)

    def set_edge_from(self, edge_id, from_node_id):
        edge_fid, old_from_node, to_node = self.edges[edge_id]
        self.edges[edge_id] = (edge_fid, from_node_id, to_node)

    def get_edge_fid(self, fid_id):
        if fid_id in self.edges:
            return self.edges[fid_id][0]
        else:
            return None

    def get_edge_from(self, edge_id):
        if edge_id in self.edges:
            return self.edges[edge_id][1]
        else:
            return None

    def replace_edge(self, old_edge_id, new_edge_fid):
        assert new_edge_fid not in self.all_fids()
        new_edge_id = new_edge_fid.get_ident()
        old_fid, from_node, to_node = self.edges[old_edge_id]
        self.edges[new_edge_id] = (new_edge_fid, from_node, to_node)
        if from_node:
            self.get_node(from_node).replace_edge(old_edge_id, new_edge_id)
        if to_node:
            self.get_node(to_node).replace_edge(old_edge_id, new_edge_id)
        del self.edges[old_edge_id]

    def get_stdin(self):
        stdin_id = self.get_stdin_id()
        stdin_fid = self.get_edge_fid(stdin_id)
        return stdin_fid

    def get_stdout(self):
        stdout_id = self.get_stdout_id()
        stdout_fid = self.get_edge_fid(stdout_id)
        return stdout_fid

    ## Gets the fid that points to the stdin of this DFG
    def get_stdin_id(self):
        ## ASSERT: There must be only one
        stdin_id = None
        for edge_id, (edge_fid, _from, _to) in self.edges.items():
            resource = edge_fid.get_resource()
            if resource.is_stdin():
                assert stdin_id is None
                stdin_id = edge_id
        return stdin_id

    def get_stdout_id(self):
        ## ASSERT: There must be only one
        stdout_id = None
        for edge_id, (edge_fid, _from, _to) in self.edges.items():
            resource = edge_fid.get_resource()
            if resource.is_stdout():
                # This is not true when using distributed_exec
                # assert(stdout_id is None)
                stdout_id = edge_id
        return stdout_id

    def serialize(self):
        output = "Nodes:\n"
        all_file_ids = ""
        for i, node in enumerate(self.nodes):
            serialized_input_file_ids = " ".join(
                [fid.serialize() for fid in node.get_input_file_ids()]
            )
            serialized_output_file_ids = " ".join(
                [fid.serialize() for fid in node.get_output_file_ids()]
            )
            all_file_ids += serialized_input_file_ids + " "
            all_file_ids += serialized_output_file_ids + " "
            output += "{} in: {} out: {} command: {}\n".format(
                i,
                serialized_input_file_ids,
                serialized_output_file_ids,
                node.serialize(),
            )
        output = "File ids:\n{}\n".format(all_file_ids) + output
        return output

    def to_ast(self, drain_streams) -> "list[AstNode]":
        asts = []

        ## Initialize the pids_to_kill variable
        asts.append(self.init_pids_to_kill())

        fileIdGen = self.get_file_id_gen()

        ## Redirect stdin
        stdin_id = self.get_stdin_id()
        if not stdin_id is None:
            ## Create a new ephemeral resource to redirect stdin to.
            fid = fileIdGen.next_file_id()
            fid.make_ephemeral()
            file_to_redirect_to = fid.to_ast()
            ## Change the stdin_id to point to this resource
            _prev_fid, from_node, to_node = self.edges[stdin_id]
            self.edges[stdin_id] = (fid, from_node, to_node)
            ## Create a command that redirects stdin to this ephemeral fid
            redirect_stdin_script = os.path.join(
                config.PASH_TOP, config.config["runtime"]["redirect_stdin_binary"]
            )
            com_args = [
                string_to_argument("source"),
                string_to_argument(redirect_stdin_script),
                file_to_redirect_to,
            ]
            com = make_command(com_args)
            asts.append(com)

        ## Make the dataflow graph
        ##
        ## TODO: Normally this should have all sink nodes at the end, but
        ##       for now we just have the stdout node in the end
        ##       (since this is always the output in our benchmarks).
        # sink_node_ids = self.sink_nodes()
        ##
        ## TODO: Support more than one output (and less than one).
        ##       For this we need to update wait.
        ##
        ## For now we just allow more than one output by waiting for one of them
        ## at random.
        stdout_edge_id = self.get_stdout_id()
        if not stdout_edge_id is None:
            sink_node_ids = [self.edges[stdout_edge_id][1]]
        else:
            sink_node_ids = self.sink_nodes()
            sink_node_ids = [sink_node_ids[0]]

        for node_id, node in self.nodes.items():
            if not node_id in sink_node_ids:
                node_ast = node.to_ast(self.edges, drain_streams)
                asts.append(make_background(node_ast))
                ## Gather all pids
                assignment = self.collect_pid_assignment()
                asts.append(assignment)

        ## Put the output node in the end for wait to work.
        for node_id in sink_node_ids:
            node = self.get_node(node_id)
            node_ast = node.to_ast(self.edges, drain_streams)
            asts.append(make_background(node_ast))
            ## Gather all pids
            assignment = self.collect_pid_assignment()
            asts.append(assignment)

        ## TODO: Ideally we would like to make them as typed nodes already
        class_asts = [to_ast_node(ast_node_to_untyped_deep(ast)) for ast in asts]
        return class_asts

    def collect_pid_assignment(self):
        ## Creates:
        ## pids_to_kill="$! $pids_to_kill"
        var_name = "pids_to_kill"
        rval = quote_arg(
            [standard_var_ast("!"), char_to_arg_char(" "), standard_var_ast(var_name)]
        )
        return make_assignment(var_name, [rval])

    def init_pids_to_kill(self):
        ## Creates:
        ## pids_to_kill=""
        var_name = "pids_to_kill"
        rval = quote_arg([])
        return make_assignment(var_name, [rval])

    ## TODO: Delete this
    def set_ast(self, ast):
        self.ast = ast

    def set_background(self, background):
        self.background = background

        if background:
            ## Since the IR is in the background, we don't have access to
            ## its stdin, stdout anymore
            self.stdin = []
            self.stdout = []

    def is_in_background(self):
        return self.background

    def pipe_append(self, other):
        assert self.valid()
        assert other.valid()

        ## This combines the two IRs by adding all of the nodes
        ## together, and by union-ing the stdout of the first with the
        ## stdin of the second.
        ##
        ## Question: What happens if one of them is NULL. This
        ##           shouldn't be the case after we check that
        ##           both self and other are not empty.
        my_out = self.get_stdout_id()
        other_in = other.get_stdin_id()
        assert not my_out is None
        assert not other_in is None

        _other_in_fid, from_node, other_in_node_id = other.edges[other_in]
        assert from_node is None
        ## ... = OtherInNode(..., other_in, ...)
        ##          v
        ## ... = OtherInNode(..., my_out, ...)
        other_in_node = other.nodes[other_in_node_id]
        other_in_node.replace_edge(other_in, my_out)
        other.edges.pop(other_in)

        ## Make the my_out id to be ephemeral file.
        my_out_fid, from_node, to_node = self.edges[my_out]
        assert to_node is None
        my_out_fid.make_ephemeral()

        ## Add the other node in my edges
        self.edges[my_out] = (my_out_fid, from_node, other_in_node_id)

        ## Just call union here
        self.union(other)

    def background_union(self, other):
        assert self.valid()
        assert other.valid()
        assert self.is_in_background()
        ## This combines two IRs where at least the first one is in
        ## background. This means that the stdin only works with the second
        ## the second (or None if both are in background). Also if
        ## both are in background, their union is also in background.

        ## If one of them is not in the background, then the whole
        ## thing isn't.
        if not other.is_in_background():
            self.set_background(other.is_in_background())

        self.union(other)

    def union(self, other):
        ## Merge the nodes of the two DFGs
        all_nodes = {**self.nodes, **other.nodes}

        ## Merge edges
        all_edges = {**self.edges, **other.edges}

        ## TODO: Check that all ids are OK (no cycles etc)
        self.nodes = all_nodes
        self.edges = all_edges

        ## TODO: Handle connections of common files (pipes, etc)
        self.combine_common_files()

    ## Combines (unions) files that refer to the same resource.
    ##
    ## WARNING: This assumes that comparing file names statically
    ## (syntactically) for equality, implies semantic
    ## equality. However this is not true. There are cases where
    ## different identifiers could refer to the same file.
    ##
    ## TODO: Fix the above issue by ensuring normalized absolute names
    ##
    ## Q: Are there also cases where a same name (let's say a
    ## variable) could point to different files in different parts of
    ## the IR? Maybe it can be true if a command is run with
    ## variable assignments)
    def combine_common_files(self):
        ## For now we just unify a file if it exists exactly twice,
        ## once at the input of a node and once at the output of
        ## another node. If a file exists in several input locations,
        ## we don't unify it. Also if a file exists in more than 1
        ## input and 1 output (or more than 1 output in general) we
        ## signal an error.

        ## For all inputs of all nodes, check if they are the output
        ## of exactly one other node.
        # log("Combining files for:", self)
        for node_id1, _node1 in self.nodes.items():
            inputs_with_file_resource = [
                (id, fid)
                for id, fid in self.get_node_input_ids_fids(node_id1)
                if fid.has_file_resource()
            ]
            for id_in, fid_in in inputs_with_file_resource:
                in_resource = fid_in.get_resource()
                number_of_out_resources = 0
                for node_id2, _node2 in self.nodes.items():
                    outputs_with_file_resource = [
                        (id, fid)
                        for id, fid in self.get_node_output_ids_fids(node_id2)
                        if fid.has_file_resource()
                    ]
                    for id_out, fid_out in outputs_with_file_resource:
                        out_resource = fid_out.get_resource()
                        ## Do not combine if the ids of the edges are already the same
                        if not id_in == id_out and in_resource == out_resource:
                            number_of_out_resources += 1
                            ## They point to the same File resource so we need to unify their fids
                            self.nodes[node_id2].replace_edge(id_out, id_in)
                            self.set_edge_from(id_in, node_id2)
                            self.set_edge_from(id_out, None)

                ## Exit with an error if a file is written by more than one node.
                ##
                ## TODO: Could this ever be improved for additional performance?
                assert number_of_out_resources <= 1

    ## Returns all the file identifiers in the IR.
    def all_fids(self):
        all_fids = [fid for fid, _from_node, _to_node in self.edges.values()]
        return all_fids

    ## Returns all input fids of the IR
    def all_input_fids(self):
        all_input_fids = [
            fid for fid, from_node, _to_node in self.edges.values() if from_node is None
        ]
        return all_input_fids

    ## Returns all output fids of the IR
    def all_output_fids(self):
        all_output_fids = [
            fid for fid, _from_node, to_node in self.edges.values() if to_node is None
        ]
        return all_output_fids

    ## Returns the sources of the IR.
    ##   This includes both the nodes that have an incoming edge (file) that has no from_node,
    ##     but also nodes that have no incoming edge (generator nodes).
    def source_nodes(self):
        sources = set()
        for _edge_fid, from_node, to_node in self.edges.values():
            if from_node is None and not to_node is None:
                sources.add(to_node)
        for node_id, node in self.nodes.items():
            if len(node.get_input_list()) == 0:
                sources.add(node_id)
        return list(sources)

    def sink_nodes(self):
        sources = set()
        for _edge_fid, from_node, to_node in self.edges.values():
            if to_node is None and not from_node is None:
                sources.add(from_node)
        return list(sources)

    def get_node_inputs(self, node_id):
        input_edge_ids = self.nodes[node_id].get_input_list()
        return input_edge_ids

    def get_node_outputs(self, node_id):
        output_edge_ids = self.nodes[node_id].get_output_list()
        return output_edge_ids

    def get_next_nodes(self, node_id):
        output_edge_ids = self.get_node_outputs(node_id)
        next_nodes = []
        for edge_id in output_edge_ids:
            _fid, from_node, to_node = self.edges[edge_id]
            assert from_node == node_id
            if not to_node is None:
                next_nodes.append(to_node)
        return next_nodes

    def get_previous_nodes(self, node_id):
        input_edge_ids = self.get_node_inputs(node_id)
        previous_nodes = []
        for edge_id in input_edge_ids:
            _fid, from_node, to_node = self.edges[edge_id]
            assert to_node == node_id
            if not from_node is None:
                previous_nodes.append(from_node)
        return previous_nodes

    def get_node_input_ids_fids(self, node_id):
        node = self.get_node(node_id)
        return [
            (input_edge_id, self.edges[input_edge_id][0])
            for input_edge_id in node.get_input_list()
        ]

    def get_node_input_ids(self, node_id):
        return [fid_id for fid_id, _fid in self.get_node_input_ids_fids(node_id)]

    def get_node_input_fids(self, node_id):
        return [fid for _fid_id, fid in self.get_node_input_ids_fids(node_id)]

    def get_node_output_ids_fids(self, node_id):
        node = self.get_node(node_id)
        return [
            (output_edge_id, self.edges[output_edge_id][0])
            for output_edge_id in node.get_output_list()
        ]

    def get_node_output_ids(self, node_id):
        return [fid_id for fid_id, _fid in self.get_node_output_ids_fids(node_id)]

    def get_node_output_fids(self, node_id):
        return [fid for _fid_id, fid in self.get_node_output_ids_fids(node_id)]

    def get_node(self, node_id):
        return self.nodes[node_id]

    ## This command gets all file identifiers of the graph, and
    ## returns a fileId generator that won't clash with the existing
    ## ones.
    def get_file_id_gen(self):
        max_id = max(self.edges.keys())
        return FileIdGen(max_id)

    def remove_node(self, node_id):
        node = self.nodes.pop(node_id)
        ## Remove the node in the edges dictionary
        for in_id in node.get_input_list():
            self.set_edge_to(in_id, None)

        for out_id in node.get_output_list():
            self.set_edge_from(out_id, None)

    def add_node(self, node):
        node_id = node.get_id()
        self.nodes[node_id] = node
        ## Add the node in the edges dictionary
        for in_id in node.get_input_list():
            self.set_edge_to(in_id, node_id)

        for out_id in node.get_output_list():
            self.set_edge_from(out_id, node_id)

    def generate_ephemeral_edges(self, fileIdGen, num_of_edges):
        file_ids = [fileIdGen.next_ephemeral_file_id() for _ in range(num_of_edges)]
        self.add_edges(file_ids)
        return [edge_fid.get_ident() for edge_fid in file_ids]

    def add_edges(self, edge_fids):
        for edge_fid in edge_fids:
            self.add_edge(edge_fid)

    def add_edge(self, edge_fid):
        fid_id = edge_fid.get_ident()
        assert not fid_id in self.edges
        self.edges[fid_id] = (edge_fid, None, None)

    ## Note: We assume that the lack of nodes is an adequate condition
    ##       to check emptiness.
    def empty(self):
        return len(self.nodes) == 0

    def apply_parallelization_to_node(
        self, node_id, parallelizer, fileIdGen, fan_out, r_split_batch_size
    ):
        splitter = parallelizer.get_splitter()
        if splitter.is_splitter_round_robin():
            self.apply_round_robin_parallelization_to_node(
                node_id, parallelizer, fileIdGen, fan_out, r_split_batch_size
            )
        elif splitter.is_splitter_round_robin_with_unwrap_flag():
            self.apply_round_robin_with_unwrap_flag_parallelization_to_node(
                node_id, parallelizer, fileIdGen, fan_out, r_split_batch_size
            )
        elif splitter.is_splitter_consec_chunks():
            self.apply_consecutive_chunks_parallelization_to_node(
                node_id, parallelizer, fileIdGen, fan_out
            )
        else:
            raise UnparallelizableError("Splitter not yet implemented for command: {}".format(self.get_node(node_id=node_id).cmd_invocation_with_io_vars.cmd_name))

    def apply_round_robin_parallelization_to_node(
        self, node_id, parallelizer, fileIdGen, fan_out, r_split_batch_size
    ):
        # TODO: this control flow should move done to aggregators once we implement them;
        #  currently, this cannot be done since splitter etc. would be added...
        aggregator_spec = parallelizer.get_aggregator_spec()
        if aggregator_spec.is_aggregator_spec_adj_lines_merge():
            raise AdjLineNotImplementedError("adj_lines_merge not yet implemented in PaSh")
        elif aggregator_spec.is_aggregator_spec_adj_lines_seq():
            raise AdjLineNotImplementedError("adj_lines_seq not yet implemented in PaSh")
        elif aggregator_spec.is_aggregator_spec_adj_lines_func():
            raise AdjLineNotImplementedError("adj_lines_func not yet implemented in PaSh")
        # END of what to move

        node = self.get_node(node_id)
        # get info from node, and delete it from graph
        (
            streaming_input,
            streaming_output,
            configuration_inputs,
        ) = (
            node.get_single_streaming_input_single_output_and_configuration_inputs_of_node_for_parallelization()
        )
        original_cmd_invocation_with_io_vars = node.cmd_invocation_with_io_vars

        can_be_fused_with_prev = False
        prev_nodes = self.get_previous_nodes(node_id)
        if len(prev_nodes) == 1:
            (
                first_pred_node,
                first_pred_cmd_inv,
            ) = self.get_only_previous_node_and_only_previous_cmd_invocation(prev_nodes)
            if isinstance(first_pred_node, r_merge.RMerge):
                can_be_fused_with_prev = True

        # remove node to be parallelized
        self.remove_node(
            node_id
        )  # remove it here already as as we need to remove edge end points ow. to avoid disconnecting graph to avoid disconnecting graph

        if can_be_fused_with_prev:
            self.remove_node(
                prev_nodes[0]
            )  # also sets respective edge to's and from's to None
            in_mapper_ids = first_pred_cmd_inv.operand_list
        else:  # cannot be fused so introduce splitter
            # splitter
            round_robin_splitter_generator = (
                lambda input_id, output_ids: r_split.make_r_split(
                    input_id, output_ids, r_split_batch_size
                )
            )
            out_split_ids = self.introduce_splitter(
                round_robin_splitter_generator, fan_out, fileIdGen, streaming_input
            )
            in_mapper_ids = out_split_ids

        # mappers
        out_mapper_ids = self.introduce_mappers(
            fan_out,
            fileIdGen,
            in_mapper_ids,
            original_cmd_invocation_with_io_vars,
            parallelizer,
        )
        out_mapper_ids = [
            out_ids[0] for out_ids in out_mapper_ids
        ]  # since we get list of list back for potential aux info

        # aggregator
        self.introduce_aggregator_for_round_robin(
            out_mapper_ids, parallelizer, streaming_output
        )

    def apply_round_robin_with_unwrap_flag_parallelization_to_node(
        self, node_id, parallelizer, fileIdGen, fan_out, r_split_batch_size
    ):
        # round robin with unwrap flag is an inferred parallelizer which ensures that
        # the command is commutative and has an aggregator for consecutive chunks;
        # thus we can check whether we can re-open a previous "RR"-parallelization ending with `r_merge`
        node = self.get_node(node_id)
        (
            streaming_input,
            streaming_output,
            configuration_inputs,
        ) = (
            node.get_single_streaming_input_single_output_and_configuration_inputs_of_node_for_parallelization()
        )
        original_cmd_invocation_with_io_vars = node.cmd_invocation_with_io_vars

        can_be_fused_with_prev = False
        prev_nodes = self.get_previous_nodes(node_id)
        if len(prev_nodes) == 1:
            (
                first_pred_node,
                first_pred_cmd_inv,
            ) = self.get_only_previous_node_and_only_previous_cmd_invocation(prev_nodes)
            if isinstance(first_pred_node, r_merge.RMerge):
                can_be_fused_with_prev = True

        # remove node to be parallelized
        self.remove_node(
            node_id
        )  # remove it here already as as we need to remove edge end points ow. to avoid disconnecting graph to avoid disconnecting graph

        if (
            can_be_fused_with_prev
        ):  # and node.is_commutative(): implied by how this kind of splitter is inferred
            self.remove_node(
                prev_nodes[0]
            )  # also sets respective edge to's and from's to None
            in_unwrap_ids = first_pred_cmd_inv.operand_list
            out_unwrap_ids = self.introduce_unwraps(fileIdGen, in_unwrap_ids)
            in_mapper_ids = out_unwrap_ids
        else:
            # splitter
            round_robin_with_unwrap_flag_splitter_generator = (
                lambda input_id, output_ids: r_split.make_r_split_with_unwrap_flag(
                    input_id, output_ids, r_split_batch_size
                )
            )
            out_split_ids = self.introduce_splitter(
                round_robin_with_unwrap_flag_splitter_generator,
                fan_out,
                fileIdGen,
                streaming_input,
            )
            in_mapper_ids = out_split_ids

        # mappers
        out_mapper_ids = self.introduce_mappers(
            fan_out,
            fileIdGen,
            in_mapper_ids,
            original_cmd_invocation_with_io_vars,
            parallelizer,
        )

        in_aggregator_ids = out_mapper_ids
        out_aggregator_id = streaming_output
        self.introduce_aggregators_for_consec_chunks(
            fileIdGen,
            in_aggregator_ids,
            original_cmd_invocation_with_io_vars,
            out_aggregator_id,
            parallelizer,
            streaming_output,
        )

    def apply_consecutive_chunks_parallelization_to_node(
        self, node_id, parallelizer, fileIdGen, fan_out
    ):
        # check whether we can fuse with previous node's parallelization:
        # we can do so if the previous node's parallelization is the same, and the aggregator is concatenation
        # Assumption: it suffices to check that the previous node is an aggregator node of type concatenate
        #  as this is unique for consecutive chunk parallelization (for now, this is true)
        node = self.get_node(node_id)
        (
            streaming_input,
            streaming_output,
            configuration_inputs,
        ) = (
            node.get_single_streaming_input_single_output_and_configuration_inputs_of_node_for_parallelization()
        )
        original_cmd_invocation_with_io_vars = node.cmd_invocation_with_io_vars

        can_be_fused_with_prev = False
        prev_nodes = self.get_previous_nodes(node_id)
        if len(prev_nodes) == 1:
            (
                first_pred_node,
                first_pred_cmd_inv,
            ) = self.get_only_previous_node_and_only_previous_cmd_invocation(prev_nodes)
            if first_pred_cmd_inv.is_aggregator_concatenate():
                can_be_fused_with_prev = True

        # remove node to be parallelized
        self.remove_node(
            node_id
        )  # remove it here already as as we need to remove edge end points ow. to avoid disconnecting graph to avoid disconnecting graph

        if can_be_fused_with_prev:
            self.remove_node(
                prev_nodes[0]
            )  # also sets respective edge to's and from's to None
            in_mapper_ids = first_pred_cmd_inv.operand_list
        else:  # cannot be fused so introduce splitter
            # splitter
            consec_chunks_splitter_generator = (
                lambda input_id, output_ids: pash_split.make_split_file(
                    input_id, output_ids
                )
            )
            out_split_ids = self.introduce_splitter(
                consec_chunks_splitter_generator, fan_out, fileIdGen, streaming_input
            )
            in_mapper_ids = out_split_ids

        # mappers
        out_mapper_ids = self.introduce_mappers(
            fan_out,
            fileIdGen,
            in_mapper_ids,
            original_cmd_invocation_with_io_vars,
            parallelizer,
        )

        # aggregators
        in_aggregator_ids = out_mapper_ids
        out_aggregator_id = streaming_output
        self.introduce_aggregators_for_consec_chunks(
            fileIdGen,
            in_aggregator_ids,
            original_cmd_invocation_with_io_vars,
            out_aggregator_id,
            parallelizer,
            streaming_output,
        )

    def get_only_previous_node_and_only_previous_cmd_invocation(self, prev_nodes):
        assert len(prev_nodes) > 0
        # get info about first one but also ensure that it is the only one if we fuse
        assert len(prev_nodes) == 1
        first_pred_id = prev_nodes[0]
        first_pred_node = self.get_node(first_pred_id)
        first_pred_cmd_inv = first_pred_node.cmd_invocation_with_io_vars
        return first_pred_node, first_pred_cmd_inv

    def introduce_splitter(
        self, splitter_generator, fan_out, fileIdGen, streaming_input
    ):
        out_split_ids = self.generate_ephemeral_edges(fileIdGen, fan_out)
        splitter = splitter_generator(streaming_input, out_split_ids)
        self.set_edge_to(streaming_input, splitter.get_id())
        for out_split_id in out_split_ids:
            self.set_edge_from(out_split_id, splitter.get_id())
        self.add_node(splitter)
        return out_split_ids

    def introduce_mappers(
        self,
        fan_out,
        fileIdGen,
        in_mapper_ids,
        original_cmd_invocation_with_io_vars,
        parallelizer,
    ):
        # -> [[input, aux1, aux2], [...], [...], ...]
        num_aux_mapper_to_aggregator = parallelizer.info_mapper_aggregator
        out_mapper_ids = []
        for _ in range(0, fan_out):
            out_mapper_ids.append(
                self.generate_ephemeral_edges(
                    fileIdGen, num_aux_mapper_to_aggregator + 1
                )
            )
        # TODO: Fix that we use different ones here!
        # list of output, aux_output_1, aux_output_2, ...
        zip_mapper_in_out_ids = zip(in_mapper_ids, out_mapper_ids)
        all_mappers = []
        for in_id, out_ids in zip_mapper_in_out_ids:
            # BEGIN: these 4 lines could be refactored to be a function in graph such that
            # creating end point of edges and the creation of edges is not decoupled
            out_id = out_ids[0]
            aux_out_ids = out_ids[1:]
            mapper_cmd_inv = parallelizer.get_actual_mapper(
                original_cmd_invocation_with_io_vars, in_id, out_id, aux_out_ids
            )
            mapper = DFGNode.make_simple_dfg_node_from_cmd_inv_with_io_vars(
                mapper_cmd_inv
            )
            self.set_edge_to(in_id, mapper.get_id())
            self.set_edge_from(out_id, mapper.get_id())
            for aux_out_id in aux_out_ids:
                self.set_edge_from(aux_out_id, mapper.get_id())
            # END
            splitter = parallelizer.get_splitter()
            if splitter.is_splitter_round_robin():
                mapper_r_wrapped = r_wrap.wrap_node(mapper, self.edges)
                self.set_edge_to(in_id, mapper_r_wrapped.get_id())
                self.set_edge_from(out_id, mapper_r_wrapped.get_id())
                mapper = mapper_r_wrapped
            all_mappers.append(mapper)
        for new_node in all_mappers:
            self.add_node(new_node)
        return out_mapper_ids

    def introduce_unwraps(self, fileIdGen, in_unwrap_ids):
        unwrap_to_commutative_mappers_ids = self.generate_ephemeral_edges(
            fileIdGen, len(in_unwrap_ids)
        )
        in_out_unwrap_ids = zip(in_unwrap_ids, unwrap_to_commutative_mappers_ids)
        for in_unwrap, out_unwrap in in_out_unwrap_ids:
            unwrap = r_unwrap.make_unwrap_node([in_unwrap], out_unwrap)
            self.add_node(unwrap)
            self.set_edge_to(
                in_unwrap, unwrap.get_id()
            )  # from are still (wrapped) mappers
            self.set_edge_from(
                out_unwrap, unwrap.get_id()
            )  # to will be set to mappers of current node
        in_mapper_ids = unwrap_to_commutative_mappers_ids
        return in_mapper_ids

    def introduce_aggregators_for_consec_chunks(
        self,
        fileIdGen,
        in_aggregator_ids,
        original_cmd_invocation_with_io_vars,
        out_aggregator_id,
        parallelizer,
        streaming_output,
    ):
        # in_aggregator_ids: [[input, aux1, aux2, ...], [...], [...], ...]
        if parallelizer.info_mapper_aggregator == 0:
            in_aggregator_ids = [
                in_ids[0] for in_ids in in_aggregator_ids
            ]  # since we get list of list back for potential aux info
            aggregator_spec = parallelizer.get_aggregator_spec()
            if (
                aggregator_spec.is_aggregator_spec_concatenate()
                or aggregator_spec.is_aggregator_spec_custom_n_ary()
            ):
                aggregator_cmd_inv = parallelizer.get_actual_aggregator(
                    original_cmd_invocation_with_io_vars,
                    in_aggregator_ids,
                    out_aggregator_id,
                )
                aggregator = DFGNode.make_simple_dfg_node_from_cmd_inv_with_io_vars(
                    aggregator_cmd_inv
                )
                for in_aggregator_id in in_aggregator_ids:
                    self.set_edge_to(in_aggregator_id, aggregator.get_id())
                self.set_edge_from(streaming_output, aggregator.get_id())
                all_aggregators = [aggregator]
                ## Add the merge commands in the graph
                for new_node in all_aggregators:
                    self.add_node(new_node)
            elif aggregator_spec.is_aggregator_spec_custom_2_ary():
                # TODO: we simplify and assume that every mapper produces a single output for now
                map_in_aggregator_ids = [[id] for id in in_aggregator_ids]
                # TODO: turn node into cmd_invocation_with_io_vars since this is the only thing required in this function
                self.create_generic_aggregator_tree(
                    original_cmd_invocation_with_io_vars,
                    parallelizer,
                    map_in_aggregator_ids,
                    out_aggregator_id,
                    fileIdGen,
                )
            else:
                raise UnparallelizableError("aggregator kind not yet implemented for command: {}".format(original_cmd_invocation_with_io_vars.cmd_name))
        else:  # we got auxiliary information
            assert parallelizer.core_aggregator_spec.is_aggregator_spec_custom_2_ary()
            map_in_aggregator_ids = in_aggregator_ids
            self.create_generic_aggregator_tree(
                original_cmd_invocation_with_io_vars,
                parallelizer,
                map_in_aggregator_ids,
                out_aggregator_id,
                fileIdGen,
            )

    def introduce_aggregator_for_round_robin(
        self, out_mapper_ids, parallelizer, streaming_output
    ):
        aggregator_spec = parallelizer.get_aggregator_spec()
        if aggregator_spec.is_aggregator_spec_concatenate():
            in_aggregator_ids = out_mapper_ids
            out_aggregator_id = streaming_output
            aggregator = r_merge.make_r_merge_node(in_aggregator_ids, out_aggregator_id)
            for in_aggregator_id in in_aggregator_ids:
                self.set_edge_to(in_aggregator_id, aggregator.get_id())
            self.set_edge_from(streaming_output, aggregator.get_id())
            all_aggregators = [aggregator]
            ## Add the aggregator node(s) in the graph
            for new_node in all_aggregators:
                self.add_node(new_node)
        else:
            # TODO: this is where the other cases for aggregators need to be added
            pass

    ## Replicates an edge using tee and returns the new node_id.
    def tee_edge(self, edge_id, times, fileIdGen):
        ## Assert that the edge is unplugged
        assert self.edges[edge_id][2] is None

        output_fids = [fileIdGen.next_ephemeral_file_id() for _ in range(times)]
        output_ids = [fid.get_ident() for fid in output_fids]

        ## Create the tee node
        new_node = make_tee(edge_id, output_ids)
        new_node_id = new_node.get_id()

        ## Rewire the dfg
        for edge_fid in output_fids:
            self.add_from_edge(new_node_id, edge_fid)
        self.add_node(new_node)
        self.set_edge_to(edge_id, new_node_id)

        return new_node_id

    def generate_graphviz(self):
        ## TODO: It is unclear if importing in here (instead of in general)
        ##       improves startup cost of the pash_runtime when not using graphviz.
        import graphviz

        dot = graphviz.Digraph()

        ## TODO: Could use subgraph for distribution etc

        ## First generate all nodes
        for node_id, node in self.nodes.items():
            dot = node.add_dot_node(dot, node_id)

        ## (I/O) File nodes should be boxes
        dot.attr("node", shape="box")

        ## Then generate all edges and input+output files
        for fid, from_node, to_node in self.edges.values():
            ## This means that this file is an ending or starting file
            if from_node is None or to_node is None:
                ## Sometimes some source or sink files might be ephemeral
                ## TODO: We should investigate why this happens
                if fid.has_file_resource():
                    label = fid.serialize()
                    node_id = f"file-{str(fid.get_ident())}"
                    dot.node(node_id, label)

                    if from_node is None:
                        dot.edge(node_id, str(to_node))
                    else:
                        dot.edge(str(from_node), node_id)
            else:
                dot.edge(str(from_node), str(to_node))

        return dot

    ## TODO: Also it should check that there are no unreachable edges

    def edge_node_consistency(self):
        ## Check if edges and nodes are consistent
        for edge_id, (_, from_node_id, to_node_id) in self.edges.items():
            if not from_node_id is None:
                from_node = self.get_node(from_node_id)
                if not (edge_id in from_node.get_output_list()):
                    log(
                        "Consistency Error: Edge id:",
                        edge_id,
                        "is not in the node outputs:",
                        from_node,
                    )
                    return False
            if not to_node_id is None:
                to_node = self.get_node(to_node_id)
                if not (edge_id in to_node.get_input_list()):
                    log(
                        "Consistency Error: Edge id:",
                        edge_id,
                        "is not in the node inputs:",
                        to_node,
                    )
                    return False

        for node_id, node in self.nodes.items():
            for edge_id in node.get_input_list():
                _, _, to_node_id = self.edges[edge_id]
                if not (to_node_id == node_id):
                    log(
                        "Consistency Error: The to_node_id of the input_edge:",
                        edge_id,
                        "of the node:",
                        node,
                        "is equal to:",
                        to_node_id,
                    )
                    return False
            for edge_id in node.get_output_list():
                _, from_node_id, _ = self.edges[edge_id]
                if not (from_node_id == node_id):
                    log(
                        "Consistency Error: The from_node_id of the output_edge:",
                        edge_id,
                        "of the node:",
                        node,
                        "is equal to:",
                        from_node_id,
                    )
                    return False

        return True

    ## This function checks whether an IR is valid -- that is, if it
    ## has at least one node, and stdin, stdout set to some non-null
    ## file identifiers.
    def valid(self):
        return (
            len(self.nodes) > 0
            and self.edge_node_consistency()
            and (not self.is_in_background() or (self.get_stdin() is None))
        )
        ## The following is not true. Background IRs should not have stdin, but they can have stdout.
        #   and self.get_stdout() is None)))
        ## The following is not true. A DFG might not have an stdin
        #  or (not self.is_in_background()
        #      and not self.get_stdin() is None
        #      and not self.get_stdout() is None)))

    ## This is a function that creates a reduce tree for a given node
    def create_generic_aggregator_tree(
        self,
        cmd_invocation_with_io_vars,
        parallelizer,
        input_ids_for_aggregators,
        out_aggregator_id,
        fileIdGen,
    ):
        def function_to_get_binary_aggregator(in_ids, out_ids):
            if len(out_ids) == 1:
                aggregator_cmd_inv = parallelizer.get_actual_aggregator(
                    cmd_invocation_with_io_vars, in_ids, out_ids[0]
                )
                aggregator = DFGNode.make_simple_dfg_node_from_cmd_inv_with_io_vars(
                    aggregator_cmd_inv
                )
                return aggregator
            else:
                # list has been flattened ...
                num_input_ids = len(in_ids)
                assert num_input_ids % 2 == 0
                fst_normal_input = in_ids[0]
                fst_aux_inputs_from = in_ids[1 : int(num_input_ids / 2)]
                snd_normal_input = in_ids[int(num_input_ids / 2)]
                snd_aux_inputs_from = in_ids[int(num_input_ids / 2) + 1 :]
                output_to = out_ids[0]
                aux_outputs_to = out_ids[1:]
                aggregator_cmd_inv = parallelizer.get_actual_2_ary_aggregator_with_aux(
                    fst_normal_input,
                    fst_aux_inputs_from,
                    snd_normal_input,
                    snd_aux_inputs_from,
                    output_to,
                    aux_outputs_to,
                )
                aggregator = DFGNode.make_simple_dfg_node_from_cmd_inv_with_io_vars(
                    aggregator_cmd_inv
                )
                return aggregator

        ## The Aggregator node takes a sequence of input ids and an output id
        all_aggregators, new_edges, final_output_id = self.create_reduce_tree(
            lambda in_ids, out_ids: function_to_get_binary_aggregator(in_ids, out_ids),
            input_ids_for_aggregators,
            fileIdGen,
        )
        ## Add the edges in the graph
        self.add_edges(new_edges)
        ## Add the merge commands in the graph
        for new_node in all_aggregators:
            self.add_node(new_node)

        ## Replace the previous final_output_id with the previous id
        node_output_edge_id = out_aggregator_id
        final_merge_node_id = self.edges[final_output_id][1]
        final_merge_node = self.get_node(final_merge_node_id)
        final_merge_node.replace_edge(final_output_id, node_output_edge_id)
        self.set_edge_from(node_output_edge_id, final_merge_node_id)
        self.set_edge_from(final_output_id, None)

    ## This function creates the reduce tree. Both input and output file
    ## ids must be lists of lists, as the input file ids and the output
    ## file ids might contain auxiliary files.
    def create_reduce_tree(self, init_func, input_ids, fileIdGen):
        tree = []
        new_edges = []
        curr_ids = input_ids
        while len(curr_ids) > 1:
            new_level, curr_ids, new_fids = self.create_reduce_tree_level(
                init_func, curr_ids, fileIdGen
            )
            tree += new_level
            new_edges += new_fids

        # Find the final output    (provided with parameter)
        final_output_id = curr_ids[0][0]

        ## Drain the final auxiliary outputs
        final_auxiliary_outputs = curr_ids[0][1:]
        drain_fids = [
            fileIdGen.next_file_id()
            for final_auxiliary_output in final_auxiliary_outputs
        ]
        for drain_fid in drain_fids:
            drain_fid.set_resource(FileResource(Arg.string_to_arg("/dev/null")))
            new_edges.append(drain_fid)
        drain_ids = [fid.get_ident() for fid in drain_fids]

        drain_cat_commands = [
            make_cat_node([final_auxiliary_output], drain_id)
            for final_auxiliary_output, drain_id in zip(
                final_auxiliary_outputs, drain_ids
            )
        ]
        return (tree + drain_cat_commands), new_edges, final_output_id

    @staticmethod
    ## This function creates a level of the reduce tree. Both input and
    ## output file ids must be lists of lists, as the input file ids and
    ## the output file ids might contain auxiliary files.
    def create_reduce_tree_level(init_func, input_ids, fileIdGen):
        if len(input_ids) % 2 == 0:
            output_ids = []
            even_input_ids = input_ids
        else:
            output_ids = [input_ids[0]]
            even_input_ids = input_ids[1:]

        new_fids = []
        level = []
        for i in range(0, len(even_input_ids), 2):
            new_out_fids = [fileIdGen.next_ephemeral_file_id() for _ in input_ids[i]]
            new_fids += new_out_fids
            new_out_ids = [fid.get_ident() for fid in new_out_fids]
            output_ids.append(new_out_ids)
            new_node = IR.create_reduce_node(
                init_func, even_input_ids[i : i + 2], new_out_ids
            )
            level.append(new_node)
        return (level, output_ids, new_fids)

    @staticmethod
    ## This function creates one node of the reduce tree
    def create_reduce_node(init_func, input_ids, output_ids):
        return init_func(flatten_list(input_ids), output_ids)

    # TODO: this is where we need to use our aggregator spec/node



"""
PaSh Preprocessor Entry Point

This module handles preprocessing of shell scripts by:
1. Parsing the shell script to ASTs
2. Replacing candidate dataflow regions with calls to PaSh runtime
3. Unparsing the transformed ASTs back to shell syntax
"""

import sys
import os
import argparse
import logging
import socket
from datetime import datetime

from shell_ast import transformation_options, ast_to_ast
from parse import parse_shell_to_asts, from_ast_objects_to_shell
from speculative import util_spec
from util import log, logging_prefix, print_time_delta


LOGGING_PREFIX = "PaSh Preprocessor: "

## Increase the recursion limit (it seems that the parser/unparser needs it for bigger scripts)
sys.setrecursionlimit(10000)
## Note: The preprocessor is very slow for very large recursive scripts

def config_from_args(pash_args):
    """Configure logging based on command-line arguments."""
    if pash_args.log_file == "":
        logging.basicConfig(format="%(message)s")
    else:
        logging.basicConfig(
            format="%(message)s",
            filename=f"{os.path.abspath(pash_args.log_file)}",
            filemode="w",
        )

    if pash_args.debug == 0:
        logging.getLogger().setLevel(logging.ERROR)
    elif pash_args.debug == 1:
        logging.getLogger().setLevel(logging.WARNING)
    elif pash_args.debug == 2:
        logging.getLogger().setLevel(logging.INFO)
    elif pash_args.debug >= 3:
        logging.getLogger().setLevel(logging.DEBUG)


class Parser(argparse.ArgumentParser):
    """Command-line argument parser for the preprocessor."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.add_argument(
            "-d",
            "--debug",
            type=int,
            help="configure debug level; defaults to 0",
            default=1,
        )
        self.add_argument(
            "--log_file",
            help="configure where to write the log; defaults to stderr.",
            default="",
        )
        self.add_argument(
            "input",
            help="the script file to be preprocessed",
        )
        self.add_argument(
            "--output",
            help="path where the preprocessed script will be saved",
            required=True,
        )
        self.add_argument(
            "--bash",
            help="(experimental) interpret the input as a bash script file",
            action="store_true",
        )
        self.add_argument(
            "--speculative",
            help="(experimental) use the speculative execution preprocessing and runtime",
            action="store_true",
            default=False,
        )

        self.set_defaults(preprocess_mode="pash")


@logging_prefix(LOGGING_PREFIX)
def main():
    """Main entry point for the preprocessor."""
    args = parse_args()
    input_script_path = args.input

    preprocessed_shell_script = preprocess(input_script_path, args)

    # Write the preprocessed script to the output file
    fname = args.output
    log("Preprocessed script stored in:", fname)
    with open(fname, "wb") as new_shell_file:
        preprocessed_shell_script = preprocessed_shell_script.encode(
            "utf-8", errors="replace"
        )
        new_shell_file.write(preprocessed_shell_script)

    log("-" * 40)  # log end marker


def preprocess(input_script_path, args):
    """
    Preprocess a shell script.

    This function:
    1. Parses the shell script to ASTs
    2. Preprocesses ASTs by replacing candidate dataflow regions
    3. Unparses the ASTs back to shell syntax

    Args:
        input_script_path: Path to the input shell script
        args: Parsed command-line arguments

    Returns:
        The preprocessed shell script as a string
    """
    # 1. Parse shell to AST
    preprocessing_parsing_start_time = datetime.now()
    ast_objects = parse_shell_to_asts(input_script_path, bash_mode=args.bash)
    preprocessing_parsing_end_time = datetime.now()
    print_time_delta(
        "Preprocessing -- Parsing",
        preprocessing_parsing_start_time,
        preprocessing_parsing_end_time,
    )

    # 2. Preprocess ASTs by replacing candidates with calls to PaSh runtime
    preprocessing_pash_start_time = datetime.now()
    preprocessed_asts = preprocess_asts(ast_objects, args)
    preprocessing_pash_end_time = datetime.now()
    print_time_delta(
        "Preprocessing -- PaSh",
        preprocessing_pash_start_time,
        preprocessing_pash_end_time,
    )

    # 3. Unparse the ASTs back to shell syntax
    preprocessing_unparsing_start_time = datetime.now()
    preprocessed_shell_script = from_ast_objects_to_shell(preprocessed_asts)
    preprocessing_unparsing_end_time = datetime.now()
    print_time_delta(
        "Preprocessing -- Unparsing",
        preprocessing_unparsing_start_time,
        preprocessing_unparsing_end_time,
    )

    return preprocessed_shell_script


def preprocess_asts(ast_objects, args):
    """
    Preprocess AST objects based on the transformation mode.

    Args:
        ast_objects: List of parsed AST objects
        args: Parsed command-line arguments

    Returns:
        List of preprocessed AST objects
    """
    trans_mode = transformation_options.TransformationType(args.preprocess_mode)

    if trans_mode is transformation_options.TransformationType.SPECULATIVE:
        trans_options = transformation_options.SpeculativeTransformationState(
            po_file=args.partial_order_file
        )
        util_spec.initialize(trans_options)
    elif trans_mode is transformation_options.TransformationType.AIRFLOW:
        trans_options = transformation_options.AirflowTransformationState()
    else:
        trans_options = transformation_options.TransformationState()

    # Preprocess ASTs by replacing regions with calls to PaSh runtime
    preprocessed_asts = ast_to_ast.replace_ast_regions(ast_objects, trans_options)

    # For speculative mode, finalize the partial order file
    if trans_mode is transformation_options.TransformationType.SPECULATIVE:
        util_spec.serialize_partial_order(trans_options)

        # Inform the scheduler that the partial order file is ready
        unix_socket_file = os.getenv("PASH_SPEC_SCHEDULER_SOCKET")
        msg = util_spec.scheduler_server_init_po_msg(
            trans_options.get_partial_order_file()
        )
        _unix_socket_send_and_forget(unix_socket_file, msg)

    return preprocessed_asts


def _unix_socket_send_and_forget(socket_file, msg):
    """Send a message to a Unix socket without waiting for a response."""
    if socket_file is None:
        return
    try:
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        sock.connect(socket_file)
        sock.sendall(msg.encode())
        sock.close()
    except Exception as e:
        log(f"Warning: Failed to send message to scheduler socket: {e}")


def parse_args():
    """Parse command-line arguments."""
    prog_name = sys.argv[0]
    if "PASH_FROM_SH" in os.environ:
        prog_name = os.environ["PASH_FROM_SH"]
    parser = Parser(prog_name)
    args = parser.parse_args()
    config_from_args(args)

    # Configure speculative mode if enabled
    if args.speculative:
        log("PaSh is running in speculative mode...")
        args.__dict__["preprocess_mode"] = "spec"
        args.__dict__["partial_order_file"] = util_spec.partial_order_file_path()
        log(" -- Its partial order file will be stored in:", args.partial_order_file)

    # Log all arguments
    log("Arguments:")
    for arg_name, arg_val in vars(args).items():
        log(arg_name, arg_val)
    log("-" * 40)

    return args


if __name__ == "__main__":
    main()



import os
import config
import subprocess
import sys

from shasta.ast_node import ast_node_to_untyped_deep
from shasta.json_to_ast import to_ast_node
from shasta.ast_node import string_of_arg
from shasta.bash_to_shasta_ast import to_ast_node as bash_to_shasta_ast

from util import log, UnparsedScript

import libdash.parser
import libbash


## Parses straight a shell script to an AST
## through python without calling it as an executable


def parse_shell_to_asts(input_script_path, bash_mode=False):
    if bash_mode:
        return parse_shell_to_asts_bash(input_script_path)
    else:
        return parse_shell_to_asts_dash(input_script_path)


INITIALIZE_LIBDASH = True


def parse_shell_to_asts_dash(input_script_path):
    global INITIALIZE_LIBDASH
    try:
        new_ast_objects = libdash.parser.parse(input_script_path, INITIALIZE_LIBDASH)
        INITIALIZE_LIBDASH = False
        ## Transform the untyped ast objects to typed ones
        typed_ast_objects = []
        for (
            untyped_ast,
            original_text,
            linno_before,
            linno_after,
        ) in new_ast_objects:
            typed_ast = to_ast_node(untyped_ast)
            typed_ast_objects.append(
                (typed_ast, original_text, linno_before, linno_after)
            )

        return typed_ast_objects
    except libdash.parser.ParsingException as e:
        log("Parsing error!", e)
        sys.exit(1)


def parse_shell_to_asts_bash(input_script_path):
    try:
        new_ast_objects = libbash.bash_to_ast(input_script_path, with_linno_info=True)

        ## convert the libbash AST to a shasta AST
        typed_ast_objects = []
        for (
            untyped_ast,
            original_text,
            linno_before,
            linno_after,
        ) in new_ast_objects:
            typed_ast = bash_to_shasta_ast(untyped_ast)
            typed_ast_objects.append(
                (
                    typed_ast,
                    original_text.decode("utf-8", errors="replace"),
                    linno_before,
                    linno_after,
                )
            )
        return typed_ast_objects
    except RuntimeError as e:
        log("Parsing error!", e)
        sys.exit(1)


def parse_shell_to_asts_interactive(input_script_path: str):
    return libdash.parser.parse(input_script_path)


def from_ast_objects_to_shell(asts):
    shell_list = []
    for ast in asts:
        # log("Ast:", ast)
        if isinstance(ast, UnparsedScript):
            shell_list.append(ast.text)
        else:
            shell_list.append(ast.pretty())

    shell_list = [
        x.decode("utf-8", errors="replace") if isinstance(x, bytes) else x
        for x in shell_list
    ]
    return "\n".join(shell_list) + "\n"


def from_ast_objects_to_shell_file(asts, new_shell_filename):
    script = from_ast_objects_to_shell(asts)
    with open(new_shell_filename, "w", encoding="utf-8") as new_shell_file:
        new_shell_file.write(script)


## Simply wraps the string_of_arg
def pash_string_of_arg(arg, quoted=False):
    return string_of_arg(arg, quoted)



from pash_annotations.datatypes.AccessKind import make_stream_input, make_stream_output
from pash_annotations.datatypes.CommandInvocationWithIOVars import (
    CommandInvocationWithIOVars,
)

import config
import os

from ir_defs.dfg_node import DFGNode


class RUnwrap(DFGNode):
    def __init__(
        self,
        cmd_invocation_with_io_vars,
        com_redirs=None,
        com_assignments=None,
        parallelizer_list=None,
        cmd_related_properties=None,
    ):
        com_redirs = [] if com_redirs is None else com_redirs
        com_assignments = [] if com_assignments is None else com_assignments
        super().__init__(
            cmd_invocation_with_io_vars,
            com_redirs=com_redirs,
            com_assignments=com_assignments,
            parallelizer_list=parallelizer_list,
            cmd_related_properties=cmd_related_properties,
        )


def make_unwrap_node(inputs, output):
    assert len(inputs) == 1
    input_id = inputs[0]
    access_map = {input_id: make_stream_input(), output: make_stream_output()}
    r_unwrap_bin = os.path.join(
        config.PASH_TOP, config.config["runtime"]["r_unwrap_binary"]
    )
    cmd_inv_with_io_vars = CommandInvocationWithIOVars(
        cmd_name=r_unwrap_bin,
        flag_option_list=[],
        operand_list=[],
        implicit_use_of_streaming_input=input_id,
        implicit_use_of_streaming_output=output,
        access_map=access_map,
    )
    return RUnwrap(cmd_inv_with_io_vars)



"""
AST utility functions for the PaSh preprocessor.
"""

from env_var_names import loop_iters_var, loop_iter_var
from shasta.ast_node import AstNode, ArgChar, CArgChar
from util import make_kv, unzip


class PreprocessedAST:
    """Class used by the preprocessor in ast_to_ir."""

    def __init__(
        self, ast, replace_whole, non_maximal, something_replaced=True, last_ast=False
    ):
        assert isinstance(ast, AstNode)
        self.ast = ast
        self.replace_whole = replace_whole
        self.non_maximal = non_maximal
        self.something_replaced = something_replaced
        self.last_ast = last_ast

    def should_replace_whole_ast(self):
        return self.replace_whole

    def is_non_maximal(self):
        return self.non_maximal

    def will_anything_be_replaced(self):
        return self.something_replaced

    def is_last_ast(self):
        return self.last_ast


class UnparsedScript:
    """
    Represents text that was not modified at all by preprocessing,
    and therefore does not need to be unparsed.
    """

    def __init__(self, text):
        self.text = text


# === Pattern matching for the AST ===


def format_args(args):
    formatted_args = [format_arg_chars(arg_chars) for arg_chars in args]
    return formatted_args


def format_arg_chars(arg_chars):
    chars = [format_arg_char(arg_char) for arg_char in arg_chars]
    return "".join(chars)


def format_arg_char(arg_char: ArgChar) -> str:
    return arg_char.format()


def string_to_carg_char_list(string: str) -> "list[CArgChar]":
    ret = [CArgChar(ord(char)) for char in string]
    return ret


def string_to_arguments(string):
    return [string_to_argument(word) for word in string.split(" ")]


def string_to_argument(string):
    ret = [char_to_arg_char(char) for char in string]
    return ret


def concat_arguments(arg1, arg2):
    """Arguments are simply `arg_char list` and therefore can just be concatenated."""
    return arg1 + arg2


def char_to_arg_char(char):
    return ["C", ord(char)]


def escaped_char(char):
    return ["E", ord(char)]


def standard_var_ast(string):
    return make_kv("V", ["Normal", False, string, []])


def make_arith(arg):
    return make_kv("A", arg)


def make_quoted_variable(string):
    return make_kv("Q", [standard_var_ast(string)])


def quote_arg(arg):
    return make_kv("Q", arg)


def redir_append_stderr_to_string_file(string):
    return make_kv("File", ["Append", 2, string_to_argument(string)])


def redir_stdout_to_file(arg):
    return make_kv("File", ["To", 1, arg])


def redir_file_to_stdin(arg):
    return make_kv("File", ["From", 0, arg])


def make_background(body, redirections=None):
    redirections = [] if redirections is None else redirections
    lineno = 0
    node = make_kv("Background", [lineno, body, redirections])
    return node


def make_backquote(node):
    node = make_kv("B", node)
    return node


def make_subshell(body, redirections=None):
    redirections = [] if redirections is None else redirections
    lineno = 0
    node = make_kv("Subshell", [lineno, body, redirections])
    return node


def make_command(arguments, redirections=None, assignments=None):
    redirections = [] if redirections is None else redirections
    assignments = [] if assignments is None else assignments
    lineno = 0
    node = make_kv("Command", [lineno, assignments, arguments, redirections])
    return node


def make_nop():
    return make_command([string_to_argument(":")])


def make_assignment(var, value):
    lineno = 0
    assignment = (var, value)
    assignments = [assignment]
    node = make_kv("Command", [lineno, assignments, [], []])
    return node


def make_semi_sequence(asts):
    if len(asts) == 0:
        return make_nop()

    if len(asts) == 1:
        return asts[0]
    else:
        acc = asts[-1]
        # Remove the last ast
        iter_asts = asts[:-1]
        for ast in iter_asts[::-1]:
            acc = make_kv("Semi", [ast, acc])
        return acc


def make_defun(name, body):
    lineno = 0
    node = make_kv("Defun", [lineno, name, body])
    return node


# === Make some nodes ===


def make_export_var_constant_string(var_name: str, value: str):
    node = make_export_var(var_name, string_to_argument(value))
    return node


def make_export_var(var_name: str, arg_char_list):
    """An argument is an arg_char_list."""
    arg1 = string_to_argument(f"{var_name}=")
    arguments = [string_to_argument("export"), concat_arguments(arg1, arg_char_list)]
    node = make_command(arguments)
    return node


def export_pash_loop_iters_for_current_context(all_loop_ids: "list[int]"):
    if len(all_loop_ids) > 0:
        iter_var_names = [loop_iter_var(loop_id) for loop_id in all_loop_ids]
        iter_vars = [
            standard_var_ast(iter_var_name) for iter_var_name in iter_var_names
        ]
        concatted_vars = [iter_vars[0]]
        for iter_var in iter_vars[1:]:
            concatted_vars.append(char_to_arg_char("-"))
            concatted_vars.append(iter_var)
        quoted_vars = [quote_arg(concatted_vars)]
    else:
        quoted_vars = []

    # export pash_loop_iters="$pash_loop_XXX_iter $pash_loop_YYY_iter ..."
    save_loop_iters_node = make_export_var(loop_iters_var(), quoted_vars)

    return save_loop_iters_node


def make_unset_var(var_name: str):
    arguments = [string_to_argument("unset"), string_to_argument(var_name)]
    node = make_command(arguments)
    return node


def make_increment_var(var_name: str):
    arg = string_to_argument(f"{var_name}+1")
    arith_expr = make_arith(arg)
    assignments = [[var_name, [arith_expr]]]
    node = make_command([], assignments=assignments)
    return node



import socket
import os
import time
import queue
import pickle
import json

from dspash.socket_utils import (
    SocketManager,
    encode_request,
    decode_request,
    send_msg,
    recv_msg,
)
from util import log
from dspash.ir_helper import prepare_graph_for_remote_exec, to_shell_file
from dspash.utils import read_file
import config
import copy

PORT = 65425  # Port to listen on (non-privileged ports are > 1023)


class WorkerConnection:
    def __init__(self, host, port):
        self._host = socket.gethostbyaddr(host)[2][
            0
        ]  # get ip address in case host needs resolving
        self._port = port
        self._running_processes = 0
        self._online = True
        # assume client service is running, can add a way to activate later
        try:
            self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self._socket.connect((self._host, self._port))
        except Exception as e:
            self._online = False

    def is_online(self):
        # TODO: create a ping to confirm is online
        return self._online

    def get_running_processes(self):
        # request_dict = { 'type': 'query',
        #                 'fields': ['running_processes']
        # }
        # with self.socket.connect((self.host, self.port)):
        #     self.socket.send(request)
        #     # TODO wait until the command exec finishes and run this in parallel?
        #     answer = self.socket.recv(1024)
        return self._running_processes

    def send_graph_exec_request(self, graph, shell_vars, functions) -> bool:
        request_dict = {
            "type": "Exec-Graph",
            "graph": graph,
            "functions": functions,
            "shell_variables": None,  # Doesn't seem needed for now
        }
        request = encode_request(request_dict)
        # TODO: do I need to open and close connection?
        send_msg(self._socket, request)
        # TODO wait until the command exec finishes and run this in parallel?
        response_data = recv_msg(self._socket)
        if not response_data or decode_request(response_data)["status"] != "OK":
            raise Exception(f"didn't recieved ack on request {response_data}")
        else:
            # self._running_processes += 1 #TODO: decrease in case of failure or process ended
            response = decode_request(response_data)
            return True

    def close(self):
        self._socket.send("Done")
        self._socket.close()

    def _wait_ack(self):
        confirmation = self._socket.recv(4096)
        if not confirmation or decode_request(confirmation).status != "OK":
            return False
            # log(f"Confirmation not recieved {confirmation}")
        else:
            return True

    def __str__(self):
        return f"Worker {self._host}:{self._port}"

    def host(self):
        return self._host


class WorkersManager:
    def __init__(self, workers: WorkerConnection = None):
        self.workers = [] if workers is None else workers
        self.host = socket.gethostbyname(socket.gethostname())
        self.args = copy.copy(config.pash_args)
        # Required to create a correct multi sink graph
        self.args.termination = ""

    def get_worker(self, fids=None) -> WorkerConnection:
        if not fids:
            fids = []

        best_worker = None  # Online worker with least work
        for worker in self.workers:
            if not worker.is_online():
                continue

            # Skip if any provided fid isn't available on the worker machine
            if any(map(lambda fid: not fid.is_available_on(worker.host()), fids)):
                continue

            if (
                best_worker is None
                or best_worker.get_running_processes() > worker.get_running_processes()
            ):
                best_worker = worker

        if best_worker == None:
            raise Exception("no workers online where the date is stored")

        return best_worker

    def add_worker(self, host, port):
        self.workers.append(WorkerConnection(host, port))

    def add_workers_from_cluster_config(self, config_path):
        with open(config_path, "r") as f:
            cluster_config = json.load(f)

        workers = cluster_config["workers"].values()
        for worker in workers:
            host = worker["host"]
            port = worker["port"]
            self.add_worker(host, port)

    def run(self):
        workers_manager = self
        workers_manager.add_workers_from_cluster_config(
            os.path.join(config.PASH_TOP, "cluster.json")
        )

        dspash_socket = SocketManager(os.getenv("DSPASH_SOCKET"))
        while True:
            request, conn = dspash_socket.get_next_cmd()
            if request.startswith("Done"):
                dspash_socket.close()
                break
            elif request.startswith("Exec-Graph"):
                args = request.split(":", 1)[1].strip()
                filename, declared_functions_file = args.split()

                (
                    worker_subgraph_pairs,
                    shell_vars,
                    main_graph,
                ) = prepare_graph_for_remote_exec(filename, self.get_worker)
                script_fname = to_shell_file(main_graph, self.args)
                log("Master node graph stored in ", script_fname)

                # Read functions
                log("Functions stored in ", declared_functions_file)
                declared_functions = read_file(declared_functions_file)

                # Report to main shell a script to execute
                response_msg = f"OK {script_fname}"
                dspash_socket.respond(response_msg, conn)

                # Execute subgraphs on workers
                for worker, subgraph in worker_subgraph_pairs:
                    worker.send_graph_exec_request(
                        subgraph, shell_vars, declared_functions
                    )
            else:
                raise Exception(f"Unknown request: {request}")


if __name__ == "__main__":
    WorkersManager().run()



from dspash.hdfs_file_data import get_hdfs_file_data, FileData, HDFSFileConfig
from typing import List, Tuple


def get_cmd_output(cmd: str):
    ret = subprocess.check_output(
        cmd, shell=True, universal_newlines=True, stderr=subprocess.PIPE
    )
    return ret.strip()


def _remove_prefix(s: str, prefix: str) -> str:
    if s.startswith(prefix):
        return s[len(prefix) :]
    return s


def get_datanode_dir() -> str:
    data_dir = get_cmd_output("hdfs getconf -confKey dfs.datanode.data.dir")
    data_dir = _remove_prefix(data_dir, "file://")
    return data_dir


def get_file_data(filename: str) -> FileData:
    return get_hdfs_file_data(filename)


def get_file_config(filename: str) -> HDFSFileConfig:
    filedata = get_file_data(filename)
    return HDFSFileConfig(filedata)



"""
Generic preprocessing walker for shell ASTs.

This module provides a pattern-matching based walker that separates
traversal mechanics from preprocessing policy.
"""

from __future__ import annotations
from dataclasses import dataclass, field
from typing import Callable, Any, TYPE_CHECKING

from shasta.ast_node import (
    AstNode,
    PipeNode,
    CommandNode,
    BackgroundNode,
    RedirNode,
    SubshellNode,
    SemiNode,
    AndNode,
    OrNode,
    NotNode,
    IfNode,
    ForNode,
    WhileNode,
    CaseNode,
    DefunNode,
    ArithNode,
    CondNode,
    SelectNode,
    ArithForNode,
    CoprocNode,
    TimeNode,
    GroupNode,
)

if TYPE_CHECKING:
    from shell_ast.transformation_options import AbstractTransformationState

from shell_ast.ast_util import PreprocessedAST


@dataclass
class PreprocessContext:
    """Context threaded through preprocessing traversal."""

    trans_options: Any  # AbstractTransformationState
    last_object: bool = False


@dataclass
class NodeResult:
    """Result from processing a single node."""

    ast: AstNode
    replace_whole: bool = False
    non_maximal: bool = False
    something_replaced: bool = False

    def to_preprocessed_ast(self, last_ast: bool) -> PreprocessedAST:
        """Convert to PreprocessedAST for API compatibility."""
        return PreprocessedAST(
            ast=self.ast,
            replace_whole=self.replace_whole,
            non_maximal=self.non_maximal,
            something_replaced=self.something_replaced,
            last_ast=last_ast,
        )


# Type alias for custom handlers
# Handler signature: (node, ctx, walker) -> NodeResult
NodeHandler = Callable[[AstNode, PreprocessContext, "WalkPreprocess"], NodeResult]


class WalkPreprocess:
    """
    Pattern-matching based preprocessing walker.

    Separates traversal mechanics from preprocessing policy through
    a handler-based architecture.
    """

    def __init__(self, handlers: dict[str, NodeHandler] | None = None):
        """
        Initialize the walker with optional custom handlers.

        Args:
            handlers: Dict mapping node type names (lowercase) to handler functions.
                     Handlers receive (node, ctx, walker) and return NodeResult.
        """
        self._handlers = handlers or {}

    def walk(self, node: AstNode, ctx: PreprocessContext) -> PreprocessedAST:
        """
        Walk and preprocess an AST node.

        Args:
            node: The node to preprocess
            ctx: The preprocessing context

        Returns:
            PreprocessedAST with transformed node and metadata
        """
        result = self._dispatch(node, ctx)
        return result.to_preprocessed_ast(ctx.last_object)

    def walk_close(
        self, node: AstNode, ctx: PreprocessContext
    ) -> tuple[AstNode, bool]:
        """
        Walk a node with "close" semantics - preprocess and optionally replace.

        This is used for children that cannot be part of the parent's
        dataflow region (e.g., children of control flow constructs).

        Args:
            node: The child node to preprocess
            ctx: The preprocessing context

        Returns:
            Tuple of (transformed_ast, something_replaced)
        """
        preprocessed = self.walk(node, ctx)

        if preprocessed.should_replace_whole_ast():
            final_ast = ctx.trans_options.replace_df_region(
                asts=[preprocessed.ast], disable_parallel_pipelines=ctx.last_object
            )
            return final_ast, True
        else:
            return preprocessed.ast, preprocessed.will_anything_be_replaced()

    def _dispatch(self, node: AstNode, ctx: PreprocessContext) -> NodeResult:
        """Dispatch to appropriate handler based on node type."""
        node_name = type(node).NodeName.lower()

        # Check for custom handler first
        if node_name in self._handlers:
            return self._handlers[node_name](node, ctx, self)

        # Fall back to default pattern-matching walker
        return self._default_walk(node, ctx)

    def _default_walk(self, node: AstNode, ctx: PreprocessContext) -> NodeResult:
        """Default walking behavior using pattern matching."""
        match node:
            # === Leaf replacement nodes ===
            # Note: When replace_whole=True, something_replaced must also be True
            # (this is asserted in ast_to_ast.py)
            case PipeNode():
                return NodeResult(
                    ast=node,
                    replace_whole=True,
                    non_maximal=node.is_background,
                    something_replaced=True,
                )

            case CommandNode():
                if len(node.arguments) == 0:
                    # Just an assignment, not a candidate
                    return NodeResult(ast=node, something_replaced=False)
                return NodeResult(
                    ast=node, replace_whole=True, something_replaced=True
                )

            case BackgroundNode():
                return NodeResult(
                    ast=node,
                    replace_whole=True,
                    non_maximal=True,
                    something_replaced=True,
                )

            # === Binary operators with close-node semantics ===
            case SemiNode():
                return self._walk_binary_close(
                    node, ctx, "left_operand", "right_operand"
                )

            case AndNode():
                return self._walk_binary_close(
                    node, ctx, "left_operand", "right_operand"
                )

            case OrNode():
                return self._walk_binary_close(
                    node, ctx, "left_operand", "right_operand"
                )

            # === Single-child close nodes ===
            case RedirNode():
                return self._walk_single_close(node, ctx, "node")

            case SubshellNode():
                return self._walk_single_close(node, ctx, "body")

            case NotNode():
                return self._walk_single_close(node, ctx, "body")

            case GroupNode():
                return self._walk_single_close(node, ctx, "body")

            case CoprocNode():
                return self._walk_single_close(node, ctx, "body")

            case TimeNode():
                return self._walk_single_close(node, ctx, "command")

            case SelectNode():
                return self._walk_single_close(node, ctx, "body")

            case ArithForNode():
                return self._walk_single_close(node, ctx, "action")

            # === Control flow with multiple children ===
            case WhileNode():
                return self._walk_while(node, ctx)

            case ForNode():
                return self._walk_for(node, ctx)

            case IfNode():
                return self._walk_if(node, ctx)

            case CaseNode():
                return self._walk_case(node, ctx)

            case CondNode():
                return self._walk_cond(node, ctx)

            # === No-op nodes ===
            case DefunNode():
                return NodeResult(ast=node, something_replaced=False)

            case ArithNode():
                return NodeResult(ast=node, something_replaced=False)

            case _:
                raise ValueError(f"Unknown node type: {type(node).NodeName}")

    # === Helper methods for common patterns ===

    def _walk_single_close(
        self, node: AstNode, ctx: PreprocessContext, child_attr: str
    ) -> NodeResult:
        """Walk a node with a single child using close-node semantics."""
        child = getattr(node, child_attr)
        new_child, replaced = self.walk_close(child, ctx)
        setattr(node, child_attr, new_child)
        return NodeResult(ast=node, something_replaced=replaced)

    def _walk_binary_close(
        self,
        node: AstNode,
        ctx: PreprocessContext,
        left_attr: str,
        right_attr: str,
    ) -> NodeResult:
        """Walk a binary node with close-node semantics on both children."""
        left = getattr(node, left_attr)
        right = getattr(node, right_attr)

        new_left, replaced_left = self.walk_close(left, ctx)
        new_right, replaced_right = self.walk_close(right, ctx)

        setattr(node, left_attr, new_left)
        setattr(node, right_attr, new_right)

        return NodeResult(ast=node, something_replaced=replaced_left or replaced_right)

    def _walk_while(self, node: WhileNode, ctx: PreprocessContext) -> NodeResult:
        """Walk a while node."""
        # Enter loop context
        ctx.trans_options.enter_loop()

        new_test, test_replaced = self.walk_close(node.test, ctx)
        new_body, body_replaced = self.walk_close(node.body, ctx)

        node.test = new_test
        node.body = new_body

        # Exit loop context
        ctx.trans_options.exit_loop()

        return NodeResult(
            ast=node, something_replaced=test_replaced or body_replaced
        )

    def _walk_for(self, node: ForNode, ctx: PreprocessContext) -> NodeResult:
        """Walk a for node (base behavior without loop tracking injection)."""
        # Enter loop context
        ctx.trans_options.enter_loop()

        new_body, replaced = self.walk_close(node.body, ctx)
        node.body = new_body

        # Exit loop context
        ctx.trans_options.exit_loop()

        return NodeResult(ast=node, something_replaced=replaced)

    def _walk_if(self, node: IfNode, ctx: PreprocessContext) -> NodeResult:
        """Walk an if node."""
        new_cond, cond_replaced = self.walk_close(node.cond, ctx)
        new_then, then_replaced = self.walk_close(node.then_b, ctx)

        if node.else_b is not None:
            new_else, else_replaced = self.walk_close(node.else_b, ctx)
        else:
            new_else, else_replaced = None, False

        node.cond = new_cond
        node.then_b = new_then
        node.else_b = new_else

        return NodeResult(
            ast=node,
            something_replaced=cond_replaced or then_replaced or else_replaced,
        )

    def _walk_case(self, node: CaseNode, ctx: PreprocessContext) -> NodeResult:
        """Walk a case node."""
        any_replaced = False
        new_cases = []

        for case in node.cases:
            if case.get("cbody") is not None:
                new_body, replaced = self.walk_close(case["cbody"], ctx)
                case["cbody"] = new_body
                any_replaced = any_replaced or replaced
            new_cases.append(case)

        node.cases = new_cases
        return NodeResult(ast=node, something_replaced=any_replaced)

    def _walk_cond(self, node: CondNode, ctx: PreprocessContext) -> NodeResult:
        """Walk a cond node (bash [[ ]] expression)."""
        replaced_left = False
        replaced_right = False

        if node.left is not None:
            node.left, replaced_left = self.walk_close(node.left, ctx)
        if node.right is not None:
            node.right, replaced_right = self.walk_close(node.right, ctx)

        return NodeResult(
            ast=node, something_replaced=replaced_left or replaced_right
        )



import os

from ir import IR
from util import log

## Ensure that PASH_TMP_PREFIX is set by pa.sh
assert not os.getenv("PASH_TIMESTAMP") is None
PASH_TIMESTAMP = os.getenv("PASH_TIMESTAMP")
DIR_NAME = f"pash_graphviz_{PASH_TIMESTAMP}"


def maybe_init_graphviz_dir(args):
    if not args.graphviz == "no":
        init_graphviz_dir(args)


def init_graphviz_dir(args):
    graphviz_dir_path = os.path.join(args.graphviz_dir, DIR_NAME)

    try:
        os.mkdir(graphviz_dir_path)
    except:
        print(f"Error: Graphviz dir:{graphviz_dir_path} could not be created!")
        exit(1)

    log("Created graphviz dir:", graphviz_dir_path)


def maybe_generate_graphviz(ir: IR, args, name="dfg"):
    if not args.graphviz == "no":
        generate_graphviz(ir, args, name=name)


def generate_graphviz(ir: IR, args, name="dfg"):
    ## TODO: It is unclear if importing in here (instead of in general)
    ##       improves startup cost of the pash_runtime when not using graphviz.
    import graphviz

    log("Generating graph...")
    dot = ir.generate_graphviz()

    ## The option argument of graphviz contains the format
    dot.format = args.graphviz

    graphviz_dir_path = os.path.join(args.graphviz_dir, DIR_NAME)
    dot.render(directory=graphviz_dir_path, filename=name)

    log("Saved graph visualization in:", os.path.join(graphviz_dir_path, name))



"""
Utility functions for the PaSh compiler.
"""

from datetime import timedelta
import logging
from typing import Optional, TypeVar, Union, List, Any

TType = TypeVar("TType")
import os
import tempfile

import config
from shasta.ast_node import CArgChar


# === List utilities ===


def flatten_list(lst):
    return [item for sublist in lst for item in sublist]


# === Logging utilities ===


def print_time_delta(prefix, start_time, end_time):
    """Log time delta between start and end time."""
    time_difference = (end_time - start_time) / timedelta(milliseconds=1)
    log("{} time:".format(prefix), time_difference, " ms")

def log(*args, end="\n", level=2):
    """Wrapper for logging."""
    if level == 1:
        concatted_args = " ".join([str(a) for a in list(args)])
        logging.warning(f"{config.LOGGING_PREFIX} {concatted_args}")
    elif level >= 2:
        concatted_args = " ".join([str(a) for a in list(args)])
        logging.info(f"{config.LOGGING_PREFIX} {concatted_args}")


# === File utilities ===


def ptempfile():
    fd, name = tempfile.mkstemp(dir=config.PASH_TMP_PREFIX)
    ## TODO: Get a name without opening the fd too if possible
    os.close(fd)
    return name


# === Optional/default utilities ===


def return_empty_list_if_none_else_itself(
    arg: Optional[TType],
) -> Union[TType, List[Any]]:  # list always empty
    if arg is None:
        return []
    else:
        return arg


def return_default_if_none_else_itself(arg: Optional[TType], default: TType) -> TType:
    if arg is None:
        return default
    else:
        return arg


# === AST key-value utilities ===


## This function gets a key and a value from the ast json format
def get_kv(dic):
    return (dic[0], dic[1])


def make_kv(key, val):
    return [key, val]


# === Custom compiler error codes ===


class UnparallelizableError(Exception):
    pass


class AdjLineNotImplementedError(Exception):
    pass


# to be raised in pash_compiler if a UnparallelizableError is caught at any point running the compiler
#   primarily to differentiate
#       --assert_compiler_success (exit with error only under general exceptions caught)
#       --assert_all_regions_parallelizable (exit with error when regions are found not parallelizable + general exceptions)
class NotAllRegionParallelizableError(Exception):
    pass


# === AST classes ===


class UnparsedScript:
    """
    Represents text that was not modified at all by preprocessing,
    and therefore does not need to be unparsed.
    """

    def __init__(self, text):
        self.text = text


# === AST pattern matching and construction ===


def format_arg_chars(arg_chars):
    chars = [arg_char.format() for arg_char in arg_chars]
    return "".join(chars)


def string_to_carg_char_list(string: str) -> "list[CArgChar]":
    ret = [CArgChar(ord(char)) for char in string]
    return ret


def string_to_arguments(string):
    return [string_to_argument(word) for word in string.split(" ")]


def string_to_argument(string):
    ret = [char_to_arg_char(char) for char in string]
    return ret


def char_to_arg_char(char):
    return ["C", ord(char)]


def standard_var_ast(string):
    return make_kv("V", ["Normal", False, string, []])


def make_quoted_variable(string):
    return make_kv("Q", [standard_var_ast(string)])


def quote_arg(arg):
    return make_kv("Q", arg)


def redir_append_stderr_to_string_file(string):
    return make_kv("File", ["Append", 2, string_to_argument(string)])


def redir_stdout_to_file(arg):
    return make_kv("File", ["To", 1, arg])


def redir_file_to_stdin(arg):
    return make_kv("File", ["From", 0, arg])


def make_background(body, redirections=None):
    redirections = [] if redirections is None else redirections
    lineno = 0
    node = make_kv("Background", [lineno, body, redirections])
    return node


def make_subshell(body, redirections=None):
    redirections = [] if redirections is None else redirections
    lineno = 0
    node = make_kv("Subshell", [lineno, body, redirections])
    return node


def make_command(arguments, redirections=None, assignments=None):
    redirections = [] if redirections is None else redirections
    assignments = [] if assignments is None else assignments
    lineno = 0
    node = make_kv("Command", [lineno, assignments, arguments, redirections])
    return node


def make_assignment(var, value):
    lineno = 0
    assignment = (var, value)
    assignments = [assignment]
    node = make_kv("Command", [lineno, assignments, [], []])
    return node


def make_semi_sequence(asts):
    if len(asts) == 0:
        return make_command([string_to_argument(":")])

    if len(asts) == 1:
        return asts[0]
    else:
        acc = asts[-1]
        # Remove the last ast
        iter_asts = asts[:-1]
        for ast in iter_asts[::-1]:
            acc = make_kv("Semi", [ast, acc])
        return acc


def make_defun(name, body):
    lineno = 0
    node = make_kv("Defun", [lineno, name, body])
    return node
